"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import os
from pathlib import Path

USER_HOME_DIR = str(Path.home())
WINDOWS_STEAM_LOCAL_FILE_PATH = 'C:\Program Files (x86)\Steam'
LINUX_STEAM_LOCAL_FILE_PATH = os.path.join(USER_HOME_DIR, ".steam/debian-installation")
LINUX_WINE_PREFIX_PATH = os.path.join(LINUX_STEAM_LOCAL_FILE_PATH, os.sep.join(["steamapps", "compatdata"]))

STEAM_GAME_INSTALL_PATH = os.sep.join(["steamapps", "common"])


def get_game_install_path():
    if sys.platform == "linux" or sys.platform == "linux2":
        return os.path.join(LINUX_STEAM_LOCAL_FILE_PATH, STEAM_GAME_INSTALL_PATH)
    else:
        return os.path.join(WINDOWS_STEAM_LOCAL_FILE_PATH, STEAM_GAME_INSTALL_PATH)


def get_user_data_path():
    if sys.platform == "linux" or sys.platform == "linux2":
        return os.path.join(LINUX_STEAM_LOCAL_FILE_PATH, "userdata")
    else:
        return os.path.join(WINDOWS_STEAM_LOCAL_FILE_PATH, "userdata")


def get_wine_prefix_path(game_id):
    return os.path.join(LINUX_WINE_PREFIX_PATH, os.sep.join([str(game_id), "pfx"]))



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK

"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import argcomplete
import argparse

parser = argparse.ArgumentParser(description='Tool to validate that an engine meets a set of criteria')
engine_source_group = parser.add_mutually_exclusive_group(required=True)
engine_source_group.add_argument("-n", "--name", type=str, nargs=2,
                                 help="The name of the engine as '<family-name>' '<variant-name>'. Note that each name "
                                      "must be quoted and there should be a space between the two names")
engine_source_group.add_argument('-u', "--variant-uid", type=str, help="The uid of the variant to check")
engine_source_group.add_argument('-c', "--exported-car-path", type=str,
                                 help="A path to an exported car - the engine of the car will be checked. "
                                      "Can be a full path to an exported zip file or unpacked folder,"
                                      " or the name of a zip file or unpacked folder in the BeamNG mod directory")
parser.add_argument('spec_file', help="A file containing the criteria which the engine must meet")


def main():
    import sys
    import sim_racing_tools.automation.scripts.scrutineering_impl as scrutineering_impl
    parser.set_defaults(func=scrutineering_impl.validate_engine)
    args = parser.parse_args()
    sys.exit(args.func(args))


if __name__ == '__main__':
    main()



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import ac
import acsys
import sys
import os
import platform
from collections import OrderedDict


if platform.architecture()[0] == "64bit":
    sysdir = "stdlib64"
else:
    sysdir = "stdlib"

ac.console(os.getcwd())
add_path = os.path.join(os.path.dirname(__file__), sysdir)
ac.console("Path to add {}".format(add_path))
sys.path.insert(0, add_path)
os.environ['PATH'] = os.environ['PATH'] + ";."

from third_party.sim_info import info

appWindow = 0
active_label = None
car_label = None
boost_map = OrderedDict()
app_name = "Boost Tracker"
active = False


# This function gets called by AC when the Plugin is initialised
# The function has to return a string with the plugin name
def acMain(ac_version):
    global appWindow, app_name, active_label, car_label
    appWindow = ac.newApp(app_name)
    ac.setSize(appWindow, 100, 100)
    ac.addRenderCallback(appWindow, onFormRender)
    ac.addOnAppActivatedListener(appWindow, on_activate)
    ac.addOnAppDismissedListener(appWindow, on_deactivate)
    car_label = ac.addLabel(appWindow, info.static.carModel)
    ac.setPosition(car_label, 25, 25)
    active_label = ac.addLabel(appWindow, "INACTIVE")
    ac.setPosition(active_label, 25, 50)
    return app_name


def acUpdate(delta):
    """
    This is where you update your app window ( != OpenGL graphics )
    such as : labels , listener , ect ...
    """
    global active, boost_map, active_label
    if not active or ac.getCarState(0, acsys.CS.Gas) < 1:
        ac.setText(active_label, "INACTIVE")
        return
    ac.setText(active_label, "ACTIVE")
    rpm = ac.getCarState(0, acsys.CS.RPM)
    boost = ac.getCarState(0, acsys.CS.TurboBoost)
    boost_map[rpm] = boost


def acShutdown():
    global active
    if active:
        active = False
        _write_results_file()


def on_activate(delta):
    global active, boost_map, car_label
    boost_map.clear()
    ac.setText(car_label, info.static.carModel)
    active = True


def on_deactivate(delta):
    global active, boost_map
    active = False
    _write_results_file()


def _write_results_file():
    output_dir = _get_car_directory(info.static.carModel)
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)
    with open(os.path.join(output_dir, "boost_tracker.csv"), "w+") as f:
        f.write("rpm,boost_bar\n")
        for rpm, boost in boost_map.items():
            f.write("{},{}\n".format(rpm, boost))


def _get_car_directory(car_name):
    return os.path.join(get_game_install_path(), "assettocorsa", "content", "cars", car_name, "data")


USER_HOME_DIR = os.path.expanduser("~")
WINDOWS_STEAM_LOCAL_FILE_PATH = 'C:\Program Files (x86)\Steam'
LINUX_STEAM_LOCAL_FILE_PATH = os.path.join(USER_HOME_DIR, ".steam/debian-installation")
LINUX_WINE_PREFIX_PATH = os.path.join(LINUX_STEAM_LOCAL_FILE_PATH, os.sep.join(["steamapps", "compatdata"]))

STEAM_GAME_INSTALL_PATH = os.sep.join(["steamapps", "common"])


def get_game_install_path():
    if sys.platform == "linux" or sys.platform == "linux2":
        return os.path.join(LINUX_STEAM_LOCAL_FILE_PATH, STEAM_GAME_INSTALL_PATH)
    else:
        return os.path.join(WINDOWS_STEAM_LOCAL_FILE_PATH, STEAM_GAME_INSTALL_PATH)


def get_wine_prefix_path(game_id):
    return os.path.join(LINUX_WINE_PREFIX_PATH, os.sep.join([str(game_id), "pfx"]))


def onFormRender(delta):
    pass



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import toml
import os
import csv
from collections import OrderedDict

from typing import List
from sim_racing_tools.assetto_corsa.utils import IniObj, extract_ini_primitive_value

NATURALLY_ASPIRATED = "n/a"
TURBO = "turbo"

FROM_COAST_REF = "FROM_COAST_REF"

ENGINE_INI_FILENAME = "engine.ini"
METADATA_FILENAME = "engine-metadata.toml"
BOOST_FILENAME = "boost.csv"
THERMAL_EFFICIENCY_RATIO_LUT_NAME = "therm_eff.lut"
FUEL_FLOW_LUT_NAME = "max_flow.lut"


class NoEngineIni(ValueError):
    def __init__(self, path):
        super(NoEngineIni, self).__init__(f"There is no {ENGINE_INI_FILENAME} file within {path}")


def load_engine(engine_path):
    e = Engine()
    ini_path = os.path.join(engine_path, ENGINE_INI_FILENAME)
    if not os.path.isfile(ini_path):
        raise NoEngineIni(engine_path)
    engine_ini_data = IniObj(ini_path)
    e.load_settings_from_ini(engine_ini_data)
    return e


class EngineSources(object):
    ASSETTO_CORSA = "ac"
    AUTOMATION = "automation"


class EngineUIData(object):
    @staticmethod
    def from_dict(data_dict):
        e = EngineUIData()
        if "TorqueCurve" in data_dict:
            e.torque_curve = data_dict["TorqueCurve"]
        if "PowerCurve" in data_dict:
            e.power_curve = data_dict["PowerCurve"]
        if "max-torque" in data_dict:
            e.max_torque = data_dict["max-torque"]
        if "max-power" in data_dict:
            e.max_power = data_dict["max-power"]
        return e

    def __init__(self):
        self.torque_curve: List[List[str]] = list()
        self.power_curve: List[List[str]] = list()
        self.max_torque = "---"
        self.max_power = "---"

    def to_dict(self):
        return {"TorqueCurve": self.torque_curve,
                "PowerCurve": self.power_curve,
                "max-torque": self.max_torque,
                "max-power": self.max_power}


class EngineMetadata(object):
    def __init__(self):
        self.source = None
        self.ui_data = None
        self.mass_kg: int or None = None
        self.boost_curve = dict()
        self.efficiency_dict = dict()
        self.info_dict = dict()

    def load(self, data_dir):
        metadata_path = os.path.join(data_dir, METADATA_FILENAME)
        if os.path.isfile(metadata_path):
            with open(metadata_path, "r") as f:
                data_dict = toml.load(f)
                for a in ["source", "mass_kg", "info_dict"]:
                    if a in data_dict:
                        setattr(self, a, data_dict[a])
                if "ui_data" in data_dict:
                    self.ui_data = EngineUIData.from_dict(data_dict["ui_data"])
        boost_filepath = os.path.join(data_dir, BOOST_FILENAME)
        if os.path.isfile(boost_filepath):
            with open(boost_filepath, "r") as f:
                self.boost_curve = {int(row["rpm"]): float(row["boost_bar"])
                                    for row in csv.DictReader(f, delimiter=',')}

    def write(self, output_dir):
        self._write_metadate_file(output_dir)
        self._write_boost_curve(output_dir)

    def _write_metadate_file(self, output_dir):
        with open(os.path.join(output_dir, METADATA_FILENAME), "w+") as f:
            ui_data = None if not self.ui_data else self.ui_data.to_dict()
            toml.dump({"source": self.source,
                       "mass_kg": self.mass_kg,
                       "ui_data": ui_data,
                       "info_dict": self.info_dict}, f)

    def _write_boost_curve(self, output_dir):
        if self.boost_curve:
            with open(os.path.join(output_dir, BOOST_FILENAME), "w+") as f:
                f.write("rpm,boost_bar\n")
                for rpm, boost_bar in self.boost_curve.items():
                    f.write(f'{round(rpm)},{boost_bar}\n')


class ExtendedFuelConsumptionData(object):
    def __init__(self):
        self.idle_throttle: float = 0.03
        self.idle_cutoff: int = 0
        self.mechanical_efficiency: float = 0.85

    def load_base_params_from_ini(self, ini_object):
        engine_data_params = ["MECHANICAL_EFFICIENCY", "IDLE_THROTTLE", "IDLE_CUTOFF"]
        for param in engine_data_params:
            if param in ini_object["ENGINE_DATA"]:
                setattr(self, param.lower(), ini_object["ENGINE_DATA"][param])

    def update_ini_with_base_params(self, ini_object):
        ini_object["ENGINE_DATA"]["MECHANICAL_EFFICIENCY"] = self.mechanical_efficiency
        ini_object["ENGINE_DATA"]["IDLE_THROTTLE"] = self.idle_throttle
        ini_object["ENGINE_DATA"]["IDLE_CUTOFF"] = self.idle_cutoff


class FuelConsumptionEfficiency(ExtendedFuelConsumptionData):
    def __init__(self):
        super(FuelConsumptionEfficiency, self).__init__()
        self.thermal_efficiency: float = 0.30
        self.thermal_efficiency_dict = None
        self.fuel_lhv: int = 43
        self.turbo_efficiency: float or None = None

    def load_from_ini(self, ini_object):
        self.load_base_params_from_ini(ini_object)
        if "FUEL_CONSUMPTION" in ini_object:
            fuel_consumption_params = ["THERMAL_EFFICIENCY", "FUEL_LHV", "TURBO_EFFICIENCY"]
            for param in fuel_consumption_params:
                if param in ini_object["FUEL_CONSUMPTION"]:
                    setattr(self, param.lower(), ini_object["FUEL_CONSUMPTION"][param])
            if "THERMAL_EFFICIENCY_LUT" in ini_object["FUEL_CONSUMPTION"]:
                self.thermal_efficiency_dict = OrderedDict()
                with open(os.path.join(ini_object.dirname(),
                                       ini_object["FUEL_CONSUMPTION"]["THERMAL_EFFICIENCY_LUT"])) as f:
                    for line in f.readlines():
                        data = line.strip().split("|")
                        self.thermal_efficiency_dict[data[0]] = data[1]

    def update_ini_object(self, ini_object):
        self.update_ini_with_base_params(ini_object)
        ini_object["FUEL_CONSUMPTION"]["THERMAL_EFFICIENCY"] = self.thermal_efficiency
        ini_object["FUEL_CONSUMPTION"]["FUEL_LHV"] = self.fuel_lhv
        if self.turbo_efficiency:
            ini_object["FUEL_CONSUMPTION"]["TURBO_EFFICIENCY"] = self.turbo_efficiency
        if self.thermal_efficiency_dict:
            ini_object["FUEL_CONSUMPTION"]["THERMAL_EFFICIENCY_LUT"] = THERMAL_EFFICIENCY_RATIO_LUT_NAME
            with open(os.path.join(ini_object.dirname(), THERMAL_EFFICIENCY_RATIO_LUT_NAME)) as f:
                for torque_ratio, thermal_efficiency in self.thermal_efficiency_dict.items():
                    f.write(f'{torque_ratio}|{thermal_efficiency}\n')


class FuelConsumptionFlowRate(ExtendedFuelConsumptionData):
    def __init__(self):
        super(FuelConsumptionFlowRate, self).__init__()
        self.max_fuel_flow_lut = None
        self.max_fuel_flow = 0

    def load_from_ini(self, ini_object):
        self.load_base_params_from_ini(ini_object)
        if "FUEL_CONSUMPTION" not in ini_object:
            return
        self.max_fuel_flow = ini_object["FUEL_CONSUMPTION"]["MAX_FUEL_FLOW"]
        if "MAX_FUEL_FLOW_LUT" in ini_object["FUEL_CONSUMPTION"]:
            self.max_fuel_flow_lut = dict()
            with open(os.path.join(ini_object.dirname(), ini_object["FUEL_CONSUMPTION"]["MAX_FUEL_FLOW_LUT"]), "r") as f:
                for line in f.readlines():
                    data = line.strip().split("|")
                    if len(data) > 1:
                        self.max_fuel_flow_lut[data[0]] = data[1]

    def update_ini_object(self, ini_object):
        self.update_ini_with_base_params(ini_object)
        ini_object["FUEL_CONSUMPTION"] = dict()
        ini_object["FUEL_CONSUMPTION"]["MAX_FUEL_FLOW"] = self.max_fuel_flow
        if self.max_fuel_flow_lut:
            ini_object["FUEL_CONSUMPTION"]["MAX_FUEL_FLOW_LUT"] = FUEL_FLOW_LUT_NAME
            with open(os.path.join(ini_object.dirname(), FUEL_FLOW_LUT_NAME), "w+") as f:
                for rpm, max_flow_rate in self.max_fuel_flow_lut.items():
                    f.write(f'{rpm}|{max_flow_rate}\n')


class Engine(object):
    def __init__(self):
        self.metadata: EngineMetadata = EngineMetadata()
        self.ini_data = None
        self.version = 1  # The version of the assetto corsa ini file to output
        self.basic_fuel_consumption: None or int = None
        self.power_info: Power = Power()
        self.coast_curve: CoastCurve = CoastCurve()
        self.altitude_sensitivity = 0.1
        self.inertia: float = 0.0  # engine.jbeam mainEngine.inertia
        self.limiter: int = 0  # sql Variants.MaxRPM
        self.limiter_hz: int = 30
        self.minimum: int = 0  # sql Variants.IdleSpeed
        self.extended_fuel_consumption: FuelConsumptionEfficiency or FuelConsumptionFlowRate or None = None

        # We generate this from generate_resource_summary as power.lut
        self.torque_curve = None
        self.turbo: Turbo = Turbo()

        self.rpm_threshold = 0  # RPM at which the engine starts to take damage
        self.rpm_damage_k = 1  # amount of damage per second per (max - threshold)

    def load_from_dir(self, dir_name):
        self.load_settings_from_ini(IniObj(os.path.join(dir_name, "engine.ini")))

    def load_settings_from_ini(self, ini_data):
        self.metadata.load(ini_data.dirname())
        self.ini_data = ini_data
        self.version = ini_data["HEADER"]["VERSION"]
        self.altitude_sensitivity = extract_ini_primitive_value(ini_data["ENGINE_DATA"]["ALTITUDE_SENSITIVITY"], float)
        self.inertia = extract_ini_primitive_value(ini_data["ENGINE_DATA"]["INERTIA"], float)
        self.limiter = extract_ini_primitive_value(ini_data["ENGINE_DATA"]["LIMITER"], int)
        self.limiter_hz = extract_ini_primitive_value(ini_data["ENGINE_DATA"]["LIMITER_HZ"], int)
        self.minimum = extract_ini_primitive_value(ini_data["ENGINE_DATA"]["MINIMUM"], int)
        self.rpm_threshold = extract_ini_primitive_value(ini_data["DAMAGE"]["RPM_THRESHOLD"], int)
        self.rpm_damage_k = extract_ini_primitive_value(ini_data["DAMAGE"]["RPM_DAMAGE_K"], int)
        self.power_info.load_from_lut(os.path.join(self.ini_data.dirname(),
                                      extract_ini_primitive_value(self.ini_data["HEADER"]["POWER_CURVE"])))
        self.coast_curve.load_from_ini(ini_data)
        self.turbo.load_from_ini(ini_data)
        if "FUEL_CONSUMPTION" in ini_data:
            if "MAX_FUEL_FLOW" in ini_data["FUEL_CONSUMPTION"]:
                self.extended_fuel_consumption = FuelConsumptionFlowRate()
            else:
                self.extended_fuel_consumption = FuelConsumptionEfficiency()
            self.extended_fuel_consumption.load_from_ini(ini_data)

    def write(self, output_path=None, use_csp_extended_physics=False):
        if output_path is None and self.ini_data is None:
            raise IOError("No output file specified")
        ini_data = IniObj(os.path.join(output_path, "engine.ini")) if output_path else self.ini_data
        self.metadata.write(ini_data.dirname())
        if "HEADER" not in ini_data:
            ini_data["HEADER"] = dict()
        ini_data["HEADER"]["VERSION"] = self.version
        if "ENGINE_DATA" not in ini_data:
            ini_data["ENGINE_DATA"] = dict()
        ini_data["ENGINE_DATA"]["ALTITUDE_SENSITIVITY"] = self.altitude_sensitivity
        ini_data["ENGINE_DATA"]["INERTIA"] = self.inertia
        ini_data["ENGINE_DATA"]["LIMITER"] = self.limiter
        ini_data["ENGINE_DATA"]["LIMITER_HZ"] = self.limiter_hz
        ini_data["ENGINE_DATA"]["MINIMUM"] = self.minimum
        self.power_info.write_power_files(ini_data)
        self.coast_curve.update_ini(ini_data)
        if "DAMAGE" not in ini_data:
            ini_data["DAMAGE"] = dict()
        ini_data["DAMAGE"]["RPM_THRESHOLD"] = self.rpm_threshold
        ini_data["DAMAGE"]["RPM_DAMAGE_K"] = self.rpm_damage_k
        self.turbo.update_ini(ini_data)
        if use_csp_extended_physics:
            if self.extended_fuel_consumption:
                self.extended_fuel_consumption.update_ini_object(ini_data)
        else:
            for param in ["MECHANICAL_EFFICIENCY", "IDLE_THROTTLE", "IDLE_CUTOFF"]:
                if param in ini_data["ENGINE_DATA"]:
                    ini_data["ENGINE_DATA"].pop(param)
            if "FUEL_CONSUMPTION" in ini_data:
                ini_data.pop("FUEL_CONSUMPTION")
        ini_data.write()

    def aspiration(self):
        if self.turbo.is_present():
            return TURBO
        return NATURALLY_ASPIRATED


class Power(object):
    def __init__(self):
        self.rpm_curve: List[int] = list()
        self.torque_curve: List[int] = list()

    def load_from_lut(self, lut_path):
        with open(lut_path, "r") as f:
            for line in f.readlines():
                if "|" not in line:
                    continue
                rpm, torque = tuple(line.strip().split("|"))
                self.rpm_curve.append(int(rpm))
                self.torque_curve.append(int(torque))

    def write_power_files(self, ini_data):
        with open(os.path.join(ini_data.dirname(), "power.lut"), "w+") as f:
            for idx, rpm in enumerate(self.rpm_curve):
                f.write(f"{rpm}|{self.torque_curve[idx]}\n")
        ini_data["HEADER"]["POWER_CURVE"] = "power.lut"


class CoastCurve(object):
    """
    From https://en.wikipedia.org/wiki/Engine_efficiency
    An engine has many moving parts that produce friction. Some of these friction forces remain constant
    (as long as applied load is constant); some of these friction losses increase as engine speed increases,
    such as piston side forces and connecting bearing forces (due to increased inertia forces from the oscillating piston).
    A few friction forces decrease at higher speed, such as the friction force on the cam's lobes used to operate
    the inlet and outlet valves (the valves' inertia at high speed tends to pull the cam follower away from the cam lobe).
    Along with friction forces, an operating engine has pumping losses, which is the work required to move air into
    and out of the cylinders. This pumping loss is minimal at low speed, but increases approximately as the square of the
    speed, until at rated power an engine is using about 20% of total power production to overcome
    friction and pumping losses.
    """
    def __init__(self):
        # TODO this assumes coast is defined by COAST_REF
        self.curve_data_source: str = ""
        self.reference_rpm: int = 0
        self.torque: int = 0
        self.non_linearity = 0

    def curve_type(self):
        return self.curve_data_source

    def load_from_ini(self, ini_data):
        self.curve_data_source = extract_ini_primitive_value(ini_data["HEADER"]["COAST_CURVE"])
        if self.curve_data_source != FROM_COAST_REF:
            raise NotImplementedError("Can only handle coast data from COAST_REF section")
        self.reference_rpm = extract_ini_primitive_value(ini_data["COAST_REF"]["RPM"], int)
        self.torque = extract_ini_primitive_value(ini_data["COAST_REF"]["TORQUE"], int)
        self.non_linearity = extract_ini_primitive_value(ini_data["COAST_REF"]["NON_LINEARITY"], int)

    def update_ini(self, ini_data):
        ini_data["HEADER"]["COAST_CURVE"] = self.curve_data_source
        if "COAST_REF" not in ini_data:
            ini_data["COAST_REF"] = dict()
        ini_data["COAST_REF"]["RPM"] = self.reference_rpm
        ini_data["COAST_REF"]["TORQUE"] = self.torque
        ini_data["COAST_REF"]["NON_LINEARITY"] = self.non_linearity


class Turbo(object):
    """
    # This is going to need some work. We can get a boost curve
    # from the engine curves table. We've then somehow got to
    # create a curve close to it using [TURBO_N] sections
    # in the engine.ini file.
    """
    def __init__(self):
        self.rpm_curve: List[int] = list()
        self.boost_curve: List[float] = list()
        self.sections = list()
        """ 
        the pressure on the air intake that the valve can take before opening, 
        the pressure on the intake depends on throttle, this is mostly used for fmod audio
        """
        self.pressure_threshold: None or float = None
        self.turbo_boost_threshold: float = 1.0  # level of TOTAL boost before the engine starts to take damage
        self.turbo_damage_k: int = 0  # amount of damage per second per (boost - threshold)

    def load_from_ini(self, ini_data):
        self.turbo_boost_threshold = extract_ini_primitive_value(ini_data["DAMAGE"]["TURBO_BOOST_THRESHOLD"], float)
        self.turbo_damage_k = extract_ini_primitive_value(ini_data["DAMAGE"]["TURBO_DAMAGE_K"], int)
        if "BOV" in ini_data:
            self.pressure_threshold = extract_ini_primitive_value(ini_data["BOV"]["PRESSURE_THRESHOLD"], float)
        turbo_idx = 0
        while True:
            turbo_section_name = f"TURBO_{turbo_idx}"
            if turbo_section_name not in ini_data:
                break
            section = TurboSection()
            section.load_from_ini(turbo_idx, ini_data)
            self.sections.append(section)
            turbo_idx += 1

    def is_present(self):
        return len(self.sections) > 0

    def update_ini(self, ini_data):
        ini_data["DAMAGE"]["TURBO_BOOST_THRESHOLD"] = self.turbo_boost_threshold
        ini_data["DAMAGE"]["TURBO_DAMAGE_K"] = self.turbo_damage_k
        if self.is_present():
            for turbo_idx, section in enumerate(self.sections):
                section.update_ini_with_section(turbo_idx, ini_data)
            if self.pressure_threshold:
                if "BOV" not in ini_data:
                    ini_data["BOV"] = dict()
                ini_data["BOV"]["PRESSURE_THRESHOLD"] = self.pressure_threshold
        else:
            turbo_idx = 0
            while True:
                turbo_section_name = f"TURBO_{turbo_idx}"
                if turbo_section_name not in ini_data:
                    break
                ini_data.pop(turbo_section_name)
                turbo_idx += 1
            if "BOV" in ini_data:
                ini_data.pop("BOV")


class TurboSection(object):
    """
    Example:
        [TURBO_0]
        LAG_DN=0.99				; Interpolation lag used slowing down the turbo
        LAG_UP=0.992				; Interpolation lag used to spin up the turbo
        MAX_BOOST=1.38 				; Maximum boost generated. This value is never exceeded and multiply the torque
                                    ; like T=T*(1.0 + boost), so a boost of 2 will give you 3 times the torque
                                    ; at a given rpm.
        WASTEGATE=1.10				; Max level of boost before the wastegate does its things. 0 = no wastegate
        DISPLAY_MAX_BOOST=1.10		; Value used by display apps
        REFERENCE_RPM=2000			; The reference rpm where the turbo reaches maximum boost (at max gas pedal).
        GAMMA=2.5
        COCKPIT_ADJUSTABLE=0
    """
    def __init__(self):
        self.lag_dn: float = 0.0
        self.lag_up: float = 0.0
        self.max_boost: float = 0.0
        self.wastegate: float = 0.0
        self.display_max_boost: float = 0.0
        self.reference_rpm: int = 0
        self.gamma: float = 1.0
        self.cockpit_adjustable: int = 0
        self.controllers: TurboControllers = TurboControllers()

    def load_from_ini(self, section_idx, ini_data):
        turbo_section_name = f"TURBO_{section_idx}"
        self.lag_dn = extract_ini_primitive_value(ini_data[turbo_section_name]["LAG_DN"], float)
        self.lag_up = extract_ini_primitive_value(ini_data[turbo_section_name]["LAG_UP"], float)
        self.max_boost = extract_ini_primitive_value(ini_data[turbo_section_name]["MAX_BOOST"], float)
        self.wastegate = extract_ini_primitive_value(ini_data[turbo_section_name]["WASTEGATE"], float)
        self.display_max_boost = extract_ini_primitive_value(ini_data[turbo_section_name]["DISPLAY_MAX_BOOST"], float)
        self.reference_rpm = extract_ini_primitive_value(ini_data[turbo_section_name]["REFERENCE_RPM"], int)
        self.gamma = extract_ini_primitive_value(ini_data[turbo_section_name]["GAMMA"], float)
        self.cockpit_adjustable = extract_ini_primitive_value(ini_data[turbo_section_name]["COCKPIT_ADJUSTABLE"], float)
        self.controllers.load_from_dir(ini_data.dirname(), section_idx)

    def update_ini_with_section(self, section_idx, ini_data):
        turbo_section_name = f"TURBO_{section_idx}"
        if turbo_section_name not in ini_data:
            ini_data[turbo_section_name] = dict()
        ini_data[turbo_section_name]["LAG_DN"] = self.lag_dn
        ini_data[turbo_section_name]["LAG_UP"] = self.lag_up
        ini_data[turbo_section_name]["MAX_BOOST"] = self.max_boost
        ini_data[turbo_section_name]["WASTEGATE"] = self.wastegate
        ini_data[turbo_section_name]["DISPLAY_MAX_BOOST"] = self.display_max_boost
        ini_data[turbo_section_name]["REFERENCE_RPM"] = self.reference_rpm
        ini_data[turbo_section_name]["GAMMA"] = self.gamma
        ini_data[turbo_section_name]["COCKPIT_ADJUSTABLE"] = self.cockpit_adjustable
        self.controllers.write(ini_data.dirname())


class TurboControllers(object):
    def __init__(self):
        self.index = 0
        self.ini_data = None
        self.controllers = list()

    def load_from_dir(self, dirname, index):
        ctrl_filename = os.path.join(dirname, f"ctrl_turbo{index}.ini")
        if not os.path.isfile(ctrl_filename):
            return None

        self.ini_data = IniObj(ctrl_filename)
        controller_index = 0
        while True:
            if f"CONTROLLER_{controller_index}" not in self.ini_data:
                break
            c = TurboController(controller_index)
            c.load_from_ini(self.ini_data)
            self.controllers.append(c)
            controller_index += 1

    def get_filename(self):
        return f"ctrl_turbo{self.index}.ini"

    def write(self, output_dir):
        out_file = os.path.join(output_dir, self.get_filename())

        if not len(self.controllers):
            if os.path.isfile(self.get_filename()):
                os.remove(os.path.join(output_dir, self.get_filename()))
            return

        with open(out_file, "wb+") as f:
            if not self.ini_data:
                self.ini_data = IniObj(out_file)
            for controller in self.controllers:
                controller.update_ini(self.ini_data)
            self.ini_data.write(outfile=f)


TURBO_CONTROLLER_PARAMS = ["INPUT", "COMBINATOR", "FILTER", "UP_LIMIT", "DOWN_LIMIT"]
class TurboController(object):
    """
    Details taken from:
    https://www.racedepartment.com/threads/engine-ini-turbo-configuration.174373/

    Example:
        ctrl_turbo0.ini
        [CONTROLLER_0]
        INPUT=RPMS
        COMBINATOR=ADD
        ; LUT parameters are rpm|wastegate
        LUT=(0=2.0|3500=2.0|6400=1.95|6600=1.91|6800=1.85|7000=1.82|7200=1.8|7400=1.76|7600=1.73|7800=1.69|8000=1.64|8200=1.59|8400=1.53|8600=1.48|8800=1.42|9000=1.37)
        FILTER=0.99     ; new value each physics step = filter*last_step_value+(1-filter)*lut_value
        UP_LIMIT=10000  ; Set the upper limit if multiple controllers used in combination
        DOWN_LIMIT=0.0  ; Set the lower limit if multiple controllers used in combination
    """
    def __init__(self, index):
        self.index = index
        self.input = "RPMS"
        self.combinator = "ADD"
        # TODO it might be valid for this to be a filename, if so, handle the case where this is a filename
        self.lut = OrderedDict()
        self.filter: float = 0.95
        self.up_limit = 10000
        self.down_limit = 0

    def load_from_ini(self, ini_object):
        for param in TURBO_CONTROLLER_PARAMS:
            if param in ini_object:
                setattr(self, param.lower(), ini_object[f"CONTROLLER_{self.index}"][param])
        if "LUT" in ini_object[f"CONTROLLER_{self.index}"]:
            lut_value = ini_object[f"CONTROLLER_{self.index}"]["LUT"].strip()
            if lut_value.startswith("("):
                self._load_lut_from_string(lut_value)
            else:
                self._load_lut_from_file(os.path.join(ini_object.dirname, lut_value))

    def _load_lut_from_string(self, lut_string):
        lut_string = lut_string[1:-1]  # Get rid of enclosing brackets '(' ')'
        for data in lut_string.split("|"):
            self._add_lut_entry(data)

    def _load_lut_from_file(self, lut_file_path):
        with open(lut_file_path, "r") as f:
            for line in f.readlines():
                self._add_lut_entry(line)

    def _add_lut_entry(self, lut_line):
        lut_data = lut_line.strip().split("=")
        self.lut[lut_data[0]] = lut_data[1]

    def update_ini(self, ini_object):
        section_name = f"CONTROLLER_{self.index}"
        if section_name not in ini_object:
            ini_object[section_name] = dict()
        for param in TURBO_CONTROLLER_PARAMS:
            ini_object[section_name][param] = getattr(self, param.lower())
        ini_object[section_name]["LUT"] = self.get_lut_string()

    def get_lut_string(self):
        out_string = "("
        out_string += "|".join([f"{rpm}={wastegate}" for rpm, wastegate in self.lut.items()])
        out_string += ")"
        return out_string


"""
Other formulas found on the web:

Friction calculations from:
https://www.eng-tips.com/viewthread.cfm?qid=83632

Input:
RPM = 8500
CID = 355
Stroke = 3.480

FHP = ((RPM ^ 2) * Stroke * CID) / ((13.7 / ((Stroke ^ .3333) * (RPM ^ .25))) * 350000000)
Should give you 270.8 FHP

pspeed = rpm * stroke * .16666667#
factor = 1.4105813# + 1.3602189# * COS(.0001257675# * pspeed + 3.2909413#)
ftq = factor * cid
fhp = (ftq * rpm) / 5252
should give you 249.1 FHP

FMEP calculations found online
I.P = (Pmep * L * A * N * k) / 60
I.P = Indicated Power (kW)
L and A = Stroke length and Area of cross section
N = RPM
k = 1/2 for 2 stroke and 1/4 for 4 stroke

Chen-Flynn friction correlation model
FMEP = (2pi*nr*T) / Vd

the engines attempt to suck in air when the butterfly valve is closed where a near vacuum occurs.
This is a very rough approximation based on the stackoverflow answer:

the "I have no idea what I'm doing" dog meme has never been more appropriate
W (j) = difference in pressure (pa) x volume displaced (m3)
Vacuum pressure (https://www.engineeringtoolbox.com/vacuum-converter-d_460.html)
Atmospheric pressure = 101325pa
90% vacuum = 10000pa
work_done_by_one_intake_stroke = (Atmospheric pressure - Vacuum pressure) x engine_cc

1000 rpm = 16.6 rps = 8.3 intake strokes per second
work_done@1000RPM = 8.3 x work_done_by_one_intake_stroke (Watts)
torque = (9.554140127 * watts) / RPM
"""



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import os

import pandas as pd
import plotly.graph_objects as go

import sim_racing_tools.assetto_corsa.installation as installation


def main():
    car_name = sys.argv[1]
    ac_install = installation.Installation()
    car_path = os.path.join(ac_install.get_installed_cars_path(), car_name)
    if not os.path.isdir(car_path):
        print("No such car")
        return 1

    expected_boost_df = pd.read_csv(os.path.join(car_path, "data", "boost.csv"))
    actual_boost_df = pd.read_csv(os.path.join(car_path, "data", "boost_tracker.csv"))
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=expected_boost_df['rpm'], y=expected_boost_df['boost_bar'],
                             mode='lines+markers',
                             name='Expected boost'))
    fig.add_trace(go.Scatter(x=actual_boost_df['rpm'], y=actual_boost_df['boost_bar'],
                             mode='lines+markers',
                             name='Actual boost'))
    fig.show()


if __name__ == '__main__':
    main()



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import sys

if sys.platform == "win32":
    from win32com.shell import shell, shellcon

import sys
import os
import sim_racing_tools.constants as constants
import sim_racing_tools.utils as utils

GAME_NAME = "Automation"
GAME_ID = 293760

if sys.platform == "win32":
    WINDOWS_PERSONAL_DIRECTORY = shell.SHGetFolderPath(0, shellcon.CSIDL_PERSONAL, None, 0)
else:
    pass

USER_GAME_DATA_PATH = os.sep.join(["My Games", GAME_NAME])
BEAMNG_MOD_PATH = os.sep.join(["BeamNG.drive", "mods"])
BEAMNG_LATEST_VERSION_SHORTCUT = os.sep.join(["BeamNG.drive", "latest.lnk"])
SANDBOX_DB_NAME = "Sandbox_openbeta.db"

ENGINE_JBEAM_NAME = "camso_engine.jbeam"


def get_install_dir():
    return os.path.join(constants.get_game_install_path(), GAME_NAME)


def get_user_documents_dir():
    if sys.platform == "linux" or sys.platform == "linux2":
        return os.path.join(constants.get_wine_prefix_path(GAME_ID),
                            os.sep.join(["drive_c", "users", "steamuser", "My Documents"]))
    else:
        return shell.SHGetFolderPath(0, shellcon.CSIDL_PERSONAL, None, 0)


def get_user_appdata_local():
    return shell.SHGetFolderPath(0, shellcon.CSIDL_LOCAL_APPDATA, 0, 0)


def get_userdata_path():
    return os.path.join(get_user_documents_dir(), USER_GAME_DATA_PATH)


def get_sandbox_db_path():
    return os.path.join(get_userdata_path(), SANDBOX_DB_NAME)


def get_beamng_export_paths():
    locations = [os.path.join(get_user_documents_dir(), BEAMNG_MOD_PATH)]
    latest_shortcut_path = os.path.join(get_user_appdata_local(), BEAMNG_LATEST_VERSION_SHORTCUT)
    if os.path.exists(latest_shortcut_path):
        import sim_racing_tools.utils as utils
        locations.insert(0, os.path.join(utils.read_win_shortcut(latest_shortcut_path), 'mods'))
    return locations


def get_exported_car_data_dir(beamng_mod_folder_name):
    found_path = None
    for mod_location in get_beamng_export_paths():
        potential_path = os.path.join(mod_location, beamng_mod_folder_name)
        if os.path.isdir(potential_path):
            found_path = potential_path
            break
        zip_filename = os.path.join(os.path.dirname(potential_path), os.path.basename(potential_path) + ".zip")
        if os.path.exists(zip_filename):
            utils.unzip_file(zip_filename)
            found_path = potential_path
            break

    if not found_path:
        raise ValueError(f"Can't find {beamng_mod_folder_name} in {','.join(get_beamng_export_paths())}")
    return os.path.join(found_path, os.sep.join(["vehicles", beamng_mod_folder_name]))


class Installation(object):
    def __init__(self, custom_root=None):
        self.on_linux = sys.platform == "linux" or sys.platform == "linux2"
        self.installation_root = get_install_dir() if not custom_root else custom_root



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""
import os
import toml
import glob
import sim_racing_tools.automation.installation as auto_install
import sim_racing_tools.automation.sandbox as sandbox
from sim_racing_tools.automation.car_file_decoder import CarFile

SUCCESS = 0
FAILURE = 1
ARGUMENT_ERROR = 2


def validate_engine(args):
    if not os.path.isfile(args.spec_file):
        print(f"The file {args.spec_file} doesn't exist or is inaccessible")
        return ARGUMENT_ERROR

    if args.exported_car_path:
        if os.path.isabs(args.exported_car_path):
            data_dir = os.path.join(args.exported_car_path,
                                    os.sep.join(["vehicles", os.path.basename(args.exported_car_path)]))
        else:
            data_dir = auto_install.get_exported_car_data_dir(args.exported_car_path)
        try:
            car_file = glob.glob(f"{data_dir}/*.car")[0]
        except IndexError:
            print(f"No .car file present in {data_dir}")
            return ARGUMENT_ERROR

        with open(car_file, "rb") as f:
            car = CarFile(os.path.basename(car_file), f.read())
            car.parse()
            uid = car.get_data()['Car']['Variant']['UID']
    elif args.variant_uid:
        uid = args.variant_uid
    elif args.name:
        uid = sandbox.get_engine_uid_from_name(args.name[0], args.name[1])
    else:
        print("No method for getting Variant UID provided")
        return ARGUMENT_ERROR

    engine_db_data = sandbox.get_engine_data(uid)
    try:
        spec_data = toml.load(args.spec_file)
    except TypeError as e:
        print(f"Couldn't parse spec file: {str(e)}")
        return ARGUMENT_ERROR

    for key in spec_data.keys():
        if key not in engine_db_data:
            print(f"Ignoring {key} in {args.spec_file}")
            continue
        if "min" in spec_data[key]:
            if engine_db_data[key] < spec_data[key]["min"]:
                print(f"Engine doesn't meet specifications: {key} ({engine_db_data[key]}) "
                      f"is less than configured min of {spec_data[key]['min']}")
                return FAILURE
        if "max" in spec_data[key]:
            if engine_db_data[key] > spec_data[key]["max"]:
                print(f"Engine doesn't meet specifications: {key} ({engine_db_data[key]}) "
                      f"is more than configured max of {spec_data[key]['max']}")
                return FAILURE
        if "not" in spec_data[key]:
            try:
                iterator = iter(spec_data[key]["not"])
            except TypeError:
                if engine_db_data[key] == spec_data[key]["not"]:
                    print(f"Engine doesn't meet specifications: {key} ({engine_db_data[key]}) "
                          f"is equal to {spec_data[key]['not']}")
                    return FAILURE
            else:
                if engine_db_data[key] in spec_data[key]["not"]:
                    print(f"Engine doesn't meet specifications: {key} ({engine_db_data[key]}) "
                          f"is equal to {spec_data[key]['not']}")
                    return FAILURE
        if "equals" in spec_data[key]:
            if engine_db_data[key] != spec_data[key]["equals"]:
                print(f"Engine doesn't meet specifications: {key} ({engine_db_data[key]}) "
                      f"is not equal to {spec_data[key]['equals']}")
                return FAILURE
        if "one_of" in spec_data[key]:
            if engine_db_data[key] not in spec_data[key]["one_of"]:
                print(f"Engine doesn't meet specifications: {key} ({engine_db_data[key]}) "
                      f"is not one of {', '.join(spec_data[key]['one_of'])}")
                return FAILURE
    print(f"Engine meets specifications")
    return SUCCESS



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import os
import shutil
import sim_racing_tools.assetto_corsa.installation as installation

BOOST_TRACKER_APP_NAME = "boost_pressure_analyser"


def install_boost_tracker():
    install_dir = os.path.join(installation.get_install_dir(), "apps", "python", BOOST_TRACKER_APP_NAME)
    if os.path.isdir(install_dir):
        shutil.rmtree(install_dir)
    os.mkdir(install_dir)
    shutil.copytree(os.path.join(os.path.dirname(__file__), "stdlib"), os.path.join(install_dir, "stdlib"))
    shutil.copytree(os.path.join(os.path.dirname(__file__), "stdlib64"), os.path.join(install_dir, "stdlib64"))
    shutil.copytree(os.path.join(os.path.dirname(__file__), "third_party"), os.path.join(install_dir, "third_party"))
    shutil.copy(os.path.join(os.path.dirname(__file__), BOOST_TRACKER_APP_NAME + ".py"),
                os.path.join(install_dir, BOOST_TRACKER_APP_NAME + ".py"))



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import List
from sim_racing_tools.assetto_corsa.utils import IniObj, extract_ini_primitive_value


def load_drivetrain(drivetrain_ini_path):
    d = Drivetrain()
    drivetrain_ini_data = IniObj(drivetrain_ini_path)
    d.load_settings_from_ini(drivetrain_ini_data)
    return d


class Drivetrain(object):
    def __init__(self):
        self.ini_data = None
        self.version: int = 3
        self.drive_type: str = DriveType.RWD
        self.gearbox: Gearbox = Gearbox()
        self.differential: Differential = Differential()
        self.clutch_max_torque: int = 0
        self.auto_clutch: AutoClutch = AutoClutch()
        self.autoblip: AutoBlip = AutoBlip()
        self.auto_shifter: AutoShifter = AutoShifter()
        self.downshift_protection: DownshiftProtection = DownshiftProtection()

    def load_settings_from_ini(self, drivetrain_ini_data):
        self.ini_data = drivetrain_ini_data
        self.version = extract_ini_primitive_value(drivetrain_ini_data["HEADER"]["VERSION"], int)
        self.drive_type = extract_ini_primitive_value(drivetrain_ini_data["TRACTION"]["TYPE"])
        self.clutch_max_torque = extract_ini_primitive_value(drivetrain_ini_data["CLUTCH"]["MAX_TORQUE"], int)
        self.gearbox.load_settings_from_ini(drivetrain_ini_data)
        self.differential.load_settings_from_ini(drivetrain_ini_data)
        self.auto_clutch.load_settings_from_ini(drivetrain_ini_data)
        self.autoblip.load_settings_from_ini(drivetrain_ini_data)
        self.auto_shifter.load_settings_from_ini(drivetrain_ini_data)
        self.downshift_protection.load_settings_from_ini(drivetrain_ini_data)

    def write(self, output_path=None):
        if output_path is None and self.ini_data is None:
            raise IOError("No output file specified")

        ini_data = IniObj(os.path.join(output_path, "drivetrain.ini")) if output_path else self.ini_data
        if "HEADER" not in ini_data:
            ini_data["HEADER"] = dict()
        ini_data["HEADER"]["VERSION"] = self.version
        if "TRACTION" not in ini_data:
            ini_data["TRACTION"] = dict()
        ini_data["TRACTION"]["TYPE"] = self.drive_type
        if "CLUTCH" not in ini_data:
            ini_data["CLUTCH"] = dict()
        ini_data["CLUTCH"]["MAX_TORQUE"] = self.clutch_max_torque
        self.gearbox.update_ini(ini_data)
        self.gearbox.write_gear_files(os.path.dirname(os.path.abspath(ini_data.filename)))
        self.differential.update_ini(ini_data)
        self.auto_clutch.update_ini(ini_data)
        self.autoblip.update_ini(ini_data)
        self.auto_shifter.update_ini(ini_data)
        self.downshift_protection.update_ini(ini_data)
        ini_data.write()


class DriveType(object):
    RWD = "RWD"
    FWD = "FWD"
    AWD = "AWD"


EfficiencyLookup = {DriveType.RWD: 0.85, DriveType.FWD: 0.9, DriveType.AWD: 0.75}


class Gearbox(object):
    def __init__(self):
        self.count: int = 0
        self.reverse_gear: float = 0.0
        self.gears = dict()
        self.default_gears = dict()
        self.final_gears = list()
        self.default_final_gear: float = 0.0
        ''' change up time in milliseconds '''
        self.change_up_time: int = 0
        ''' change down time in milliseconds '''
        self.change_dn_time: int = 0
        ''' Auto cutoff time for upshifts in milliseconds, 0 to disable '''
        self.auto_cutoff_time: int = 0
        ''' 1=Car supports shifter, 0=car supports only paddles '''
        self.supports_shifter: int = 0
        ''' range window additional to the precise rev matching rpm that permits gear engage. '''
        self.valid_shift_rpm_window: int = 0
        ''' 
        multiplayer for gas,brake,clutch pedals that permits gear engage on different rev matching rpm. 
        the lower the more difficult.
        '''
        self.controls_window_gain: float = 0.0
        ''' gearbox inertia. default values to 0.02 if not set '''
        self.inertia: float = 0.02

    def load_settings_from_ini(self, drivetrain_ini_data):
        self.count = extract_ini_primitive_value(drivetrain_ini_data["GEARS"]["COUNT"], int)
        self.reverse_gear = extract_ini_primitive_value(drivetrain_ini_data["GEARS"]["GEAR_R"], float)
        self.default_final_gear = extract_ini_primitive_value(drivetrain_ini_data["GEARS"]["FINAL"], float)
        for gear_num in range(1, self.count+1):
            self.default_gears[gear_num] = \
                extract_ini_primitive_value(drivetrain_ini_data["GEARS"][f"GEAR_{gear_num}"], float)
        self._lookup_gear_data(drivetrain_ini_data)

        self.change_up_time = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["CHANGE_UP_TIME"], int)
        self.change_dn_time = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["CHANGE_DN_TIME"], int)
        self.auto_cutoff_time = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["AUTO_CUTOFF_TIME"], int)
        self.supports_shifter = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["SUPPORTS_SHIFTER"], int)
        self.valid_shift_rpm_window = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["VALID_SHIFT_RPM_WINDOW"], int)
        self.controls_window_gain = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["CONTROLS_WINDOW_GAIN"], float)
        if "INERTIA" in drivetrain_ini_data["GEARBOX"]:
            self.inertia = extract_ini_primitive_value(drivetrain_ini_data["GEARBOX"]["INERTIA"], float)

    def update_ini(self, ini_object):
        if "GEARS" not in ini_object:
            ini_object["GEARS"] = dict()
        ini_object["GEARS"]["COUNT"] = self.count
        ini_object["GEARS"]["GEAR_R"] = self.reverse_gear
        ini_object["GEARS"]["FINAL"] = self.default_final_gear
        for gear_num, gear_ratio in sorted(self.default_gears.items()):
            ini_object["GEARS"][f"GEAR_{gear_num}"] = gear_ratio
        self.write_gear_files(os.path.dirname(os.path.abspath(ini_object.filename)))
        if "GEARBOX" not in ini_object:
            ini_object["GEARBOX"] = dict()
        ini_object["GEARBOX"]["CHANGE_UP_TIME"] = self.change_up_time
        ini_object["GEARBOX"]["CHANGE_DN_TIME"] = self.change_dn_time
        ini_object["GEARBOX"]["AUTO_CUTOFF_TIME"] = self.auto_cutoff_time
        ini_object["GEARBOX"]["SUPPORTS_SHIFTER"] = self.supports_shifter
        ini_object["GEARBOX"]["VALID_SHIFT_RPM_WINDOW"] = self.valid_shift_rpm_window
        ini_object["GEARBOX"]["CONTROLS_WINDOW_GAIN"] = self.controls_window_gain
        ini_object["GEARBOX"]["INERTIA"] = self.inertia

    def write_gear_files(self, output_dir):
        setup_file = os.path.join(output_dir, "setup.ini")
        setup_ini = IniObj(setup_file)
        for gear_num, gear_list in sorted(self.gears.items()):
            gear_rto_filepath = os.path.join(output_dir, f"{GEAR_LOOKUP[gear_num]}.rto")
            with open(gear_rto_filepath, "w+") as f:
                for gear in gear_list:
                    f.write(f"{gear.name}|{gear.ratio}\n")
            gear_setup_section_name = f"GEAR_{gear_num}"
            if gear_setup_section_name not in setup_ini:
                setup_ini[gear_setup_section_name] = dict()
            setup_ini[gear_setup_section_name]["RATIOS"] = os.path.basename(gear_rto_filepath)
        if self.final_gears:
            with open(os.path.join(output_dir, "final.rto"), "w+") as f:
                for gear in self.final_gears:
                    f.write(f"{gear.name}|{gear.ratio}\n")
            if "FINAL_GEAR_RATIO" not in setup_ini:
                setup_ini["FINAL_GEAR_RATIO"] = dict()
            setup_ini["FINAL_GEAR_RATIO"]["RATIOS"] = "final.rto"
        setup_ini.write()

    def _lookup_gear_data(self, ini_object):
        lookup_path = os.path.dirname(os.path.abspath(ini_object.filename))
        setup_file = os.path.join(lookup_path, "setup.ini")
        if not os.path.exists(setup_file):
            return
        ratio_files_map = dict()
        setup_ini = IniObj(setup_file)
        for idx in range(1, self.count+1):
            if f"GEAR_{idx}" not in setup_ini:
                continue
            ratio_file = extract_ini_primitive_value(setup_ini[f"GEAR_{idx}"]["RATIOS"])
            if ratio_file not in ratio_files_map:
                ratio_files_map[ratio_file] = list()
            ratio_files_map[ratio_file].append(idx)

        if "FINAL_GEAR_RATIO" in setup_ini:
            ratio_file = extract_ini_primitive_value(setup_ini["FINAL_GEAR_RATIO"]["RATIOS"])
            self.final_gears.extend(Gear.load_gears_from_file(os.path.join(lookup_path, ratio_file)))

        dir_name, dirs, files = next(os.walk(lookup_path))
        for filename in files:
            if filename not in ratio_files_map:
                continue

            full_file_path = os.path.join(dir_name, filename)
            gear_list = Gear.load_gears_from_file(full_file_path)
            for gear_num in ratio_files_map[filename]:
                if gear_num not in self.gears:
                    self.gears[gear_num] = list()
                self.gears[gear_num].extend(gear_list)


FIRST = "1st"
SECOND = "2nd"
THIRD = "3rd"
FOURTH = "4th"
FIFTH = "5th"
SIXTH = "6th"
SEVENTH = "7th"
EIGHTH = "8th"
FINAL = "final"
REVERSE = "reverse"
GEAR_LOOKUP = {1: FIRST, 2: SECOND, 3: THIRD, 4: FOURTH, 5: FIFTH, 6: SIXTH, 7: SEVENTH, 8: EIGHTH}


class Gear(object):
    @staticmethod
    def load_gears_from_file(file_path):
        with open(file_path, "r") as f:
            gear_list = list()
            for line in f.readlines():
                gear_info = line.strip().split("|")
                if len(gear_info) > 1:
                    gear_list.append(Gear(gear_info[0], float(gear_info[1])))
        return gear_list

    def __init__(self, name, ratio):
        self.name: str = name
        self.ratio: float = ratio


class Differential(object):
    def __init__(self):
        self.power: float = 0.0
        self.coast: float = 0.0
        self.preload: int = 13

    def load_settings_from_ini(self, drivetrain_ini_data):
        self.power = extract_ini_primitive_value(drivetrain_ini_data["DIFFERENTIAL"]["POWER"], float)
        self.coast = extract_ini_primitive_value(drivetrain_ini_data["DIFFERENTIAL"]["COAST"], float)
        self.preload = extract_ini_primitive_value(drivetrain_ini_data["DIFFERENTIAL"]["PRELOAD"], int)

    def update_ini(self, ini_object):
        if "DIFFERENTIAL" not in ini_object:
            ini_object["DIFFERENTIAL"] = dict()
        ini_object["DIFFERENTIAL"]["POWER"] = self.power
        ini_object["DIFFERENTIAL"]["COAST"] = self.coast
        ini_object["DIFFERENTIAL"]["PRELOAD"] = self.preload


class AutoClutch(object):
    def __init__(self):
        self.upshift_profile = None
        self.downshift_profile = None
        '''
        Use the autoclutch on gear shifts even when autoclutch is set to off. 
        Needed for cars with semiautomatic gearboxes. values 1,0
        '''
        self.use_on_changes: int = 0
        self.min_rpm: int = 0
        self.max_rpm: int = 0
        self.forced_on: int = 0

    def load_settings_from_ini(self, drivetrain_ini_data):
        up_profile = extract_ini_primitive_value(drivetrain_ini_data["AUTOCLUTCH"]["UPSHIFT_PROFILE"])
        if up_profile.lower() != "none":
            self.upshift_profile = ShiftProfile.create_from_section(drivetrain_ini_data, up_profile)
        down_profile = extract_ini_primitive_value(drivetrain_ini_data["AUTOCLUTCH"]["DOWNSHIFT_PROFILE"])
        if down_profile.lower() != "none":
            self.downshift_profile = ShiftProfile.create_from_section(drivetrain_ini_data, down_profile)
        self.use_on_changes = extract_ini_primitive_value(drivetrain_ini_data["AUTOCLUTCH"]["USE_ON_CHANGES"], int)
        self.min_rpm = extract_ini_primitive_value(drivetrain_ini_data["AUTOCLUTCH"]["MIN_RPM"], int)
        self.max_rpm = extract_ini_primitive_value(drivetrain_ini_data["AUTOCLUTCH"]["MAX_RPM"], int)
        self.forced_on = extract_ini_primitive_value(drivetrain_ini_data["AUTOCLUTCH"]["FORCED_ON"], int)

    def update_ini(self, ini_object):
        if "AUTOCLUTCH" not in ini_object:
            ini_object["AUTOCLUTCH"] = dict()
        if self.upshift_profile:
            ini_object["AUTOCLUTCH"]["UPSHIFT_PROFILE"] = self.upshift_profile.name
            self.upshift_profile.update_ini(ini_object)
        if self.downshift_profile:
            ini_object["AUTOCLUTCH"]["DOWNSHIFT_PROFILE"] = self.downshift_profile.name
            self.downshift_profile.update_ini(ini_object)
        ini_object["AUTOCLUTCH"]["USE_ON_CHANGES"] = self.use_on_changes
        ini_object["AUTOCLUTCH"]["MIN_RPM"] = self.min_rpm
        ini_object["AUTOCLUTCH"]["MAX_RPM"] = self.max_rpm
        ini_object["AUTOCLUTCH"]["FORCED_ON"] = self.forced_on


class ShiftProfile(object):
    @staticmethod
    def create_from_section(ini_data, section_name):
        profile = ShiftProfile()
        profile.name = section_name
        for idx in range(3):
            profile.points.append(extract_ini_primitive_value(ini_data[section_name][f"POINT_{idx}"], int))
        return profile

    def __init__(self):
        self.name = ""
        self.points: List[int] = list()

    def update_ini(self, ini_object):
        if self.name not in ini_object:
            ini_object[self.name] = dict()
        for idx, point in enumerate(self.points):
            ini_object[self.name][f"POINT_{idx}"] = point


class AutoBlip(object):
    def __init__(self):
        """ If =1 then it is a feature of the car and cannot be disabled """
        self.electronic: int = 0
        """
        In tatuus:
          - 0 = Time to reach full level
          - 1 = Time to start releasing gas
          - 2 = Time to reach 0 gas
        """
        self.points: List[int] = list()
        """ Gas level to be reached """
        self.level: float = 0.0

    def load_settings_from_ini(self, drivetrain_ini_data):
        self.electronic = extract_ini_primitive_value(drivetrain_ini_data["AUTOBLIP"]["ELECTRONIC"], int)
        for idx in range(3):
            self.points.append(extract_ini_primitive_value(drivetrain_ini_data["AUTOBLIP"][f"POINT_{idx}"], int))
        self.level = extract_ini_primitive_value(drivetrain_ini_data["AUTOBLIP"]["LEVEL"], float)

    def update_ini(self, ini_object):
        if "AUTOBLIP" not in ini_object:
            ini_object["AUTOBLIP"] = dict()
        ini_object["AUTOBLIP"]["ELECTRONIC"] = self.electronic
        for idx, point in enumerate(self.points):
            ini_object["AUTOBLIP"][f"POINT_{idx}"] = point
        ini_object["AUTOBLIP"]["LEVEL"] = self.level


class AutoShifter(object):
    def __init__(self):
        self.up: int = 0
        self.down: int = 0
        self.slip_threshold: float = 0.0
        self.gas_cutoff_time: float = 0.0

    def load_settings_from_ini(self, drivetrain_ini_data):
        self.up = extract_ini_primitive_value(drivetrain_ini_data["AUTO_SHIFTER"]["UP"], int)
        self.down = extract_ini_primitive_value(drivetrain_ini_data["AUTO_SHIFTER"]["DOWN"], int)
        self.slip_threshold = extract_ini_primitive_value(drivetrain_ini_data["AUTO_SHIFTER"]["SLIP_THRESHOLD"], float)
        self.gas_cutoff_time = extract_ini_primitive_value(drivetrain_ini_data["AUTO_SHIFTER"]["GAS_CUTOFF_TIME"], float)

    def update_ini(self, ini_object):
        if "AUTO_SHIFTER" not in ini_object:
            ini_object["AUTO_SHIFTER"] = dict()
        ini_object["AUTO_SHIFTER"]["UP"] = self.up
        ini_object["AUTO_SHIFTER"]["DOWN"] = self.down
        ini_object["AUTO_SHIFTER"]["SLIP_THRESHOLD"] = self.slip_threshold
        ini_object["AUTO_SHIFTER"]["GAS_CUTOFF_TIME"] = self.gas_cutoff_time


class DownshiftProtection(object):
    def __init__(self):
        self.active = 1
        ''' adds a line in the log for every missed downshift '''
        self.debug = 0
        ''' How many RPM over the limiter the car is allowed to go '''
        self.overrev = 200
        self.lock_n = 1

    def load_settings_from_ini(self, drivetrain_ini_data):
        self.active = extract_ini_primitive_value(drivetrain_ini_data["DOWNSHIFT_PROTECTION"]["ACTIVE"], int)
        self.debug = extract_ini_primitive_value(drivetrain_ini_data["DOWNSHIFT_PROTECTION"]["DEBUG"], int)
        self.overrev = extract_ini_primitive_value(drivetrain_ini_data["DOWNSHIFT_PROTECTION"]["OVERREV"], int)
        self.lock_n = extract_ini_primitive_value(drivetrain_ini_data["DOWNSHIFT_PROTECTION"]["LOCK_N"], int)

    def update_ini(self, ini_object):
        if "DOWNSHIFT_PROTECTION" not in ini_object:
            ini_object["DOWNSHIFT_PROTECTION"] = dict()
        ini_object["DOWNSHIFT_PROTECTION"]["ACTIVE"] = self.active
        ini_object["DOWNSHIFT_PROTECTION"]["DEBUG"] = self.debug
        ini_object["DOWNSHIFT_PROTECTION"]["OVERREV"] = self.overrev
        ini_object["DOWNSHIFT_PROTECTION"]["LOCK_N"] = self.lock_n



"""
Copyright (c):
2003-2010, Michael Foord
2014, Eli Courtwright, Rob Dennis
All rights reserved.
E-mails :
fuzzyman AT voidspace DOT org DOT uk
eli AT courtwright DOT org
rdennis AT gmail DOT com


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    * Neither the names of Michael Foord, Eli Courtwright or Rob Dennis,
      nor the name of Voidspace,  may be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
Copyright (c):
2021 zephyrj 
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import os
import re
import six

import configobj
from configobj import BOM_LIST, Section, BOM_UTF8, match_utf8, unrepr


class IniObj(configobj.ConfigObj):
    # this regexp is taken from the original ConfigObj class and tweaked to treat
    # semi colons as comments
    _valueexp = re.compile(r'''^
            (?:
                (?:
                    (
                        (?:
                            (?:
                                (?:".*?")|              # double quotes
                                (?:'.*?')|              # single quotes
                                (?:[^'",\#;][^,\#;]*?)    # unquoted
                            )
                            \s*,\s*                     # comma
                        )*      # match all list items ending in a comma (if any)
                    )
                    (
                        (?:".*?")|                      # double quotes
                        (?:'.*?')|                      # single quotes
                        (?:[^'",\#;\s][^,]*?)|           # unquoted
                        (?:(?<!,))                      # Empty value
                    )?          # last item in a list - or string value
                )|
                (,)             # alternatively a single comma - empty list
            )
            \s*([\#;].*)?          # optional comment
            $''', re.VERBOSE)

    def __init__(self, *args, **kwargs):
        configobj.ConfigObj.__init__(self, *args, **kwargs)
        self.is_new_file = True if not os.path.isfile(self.filename) else False
        self.comment_indent = "    "
        self._valueexp = IniObj._valueexp

    def _write_line(self, indent_string, entry, this_entry, comment):
        """Write an individual line, for the write method"""
        if not self.unrepr:
            val = self._decode_element(self._quote(this_entry))
        else:
            val = repr(this_entry)

        return '%s%s%s%s%s' % (indent_string,
                               self._decode_element(self._quote(entry, multiline=False)),
                               self._a_to_u('='),
                               val,
                               self._decode_element(comment))

    def _handle_comment(self, comment):
        """Deal with a comment."""
        if not comment:
            return ''
        start = self.comment_indent
        if not comment.startswith(';'):
            start += self._a_to_u(' ; ')
        return (start + comment)

    def write(self, outfile=None, section=None):
        if self.indent_type is None:
            # this can be true if initialised from a dictionary
            self.indent_type = configobj.DEFAULT_INDENT_TYPE

        out = []
        cs = (self._a_to_u('#'), self._a_to_u(';'))
        csp = self._a_to_u('; ')
        if section is None:
            int_val = self.interpolation
            self.interpolation = False
            section = self
            for line in self.initial_comment:
                line = self._decode_element(line)
                stripped_line = line.strip()
                if stripped_line and not stripped_line.startswith(cs):
                    line = csp + line
                out.append(line)

        indent_string = self.indent_type * section.depth
        for entry in (section.scalars + section.sections):
            if entry in section.defaults:
                # don't write out default values
                continue
            for comment_line in section.comments[entry]:
                comment_line = self._decode_element(comment_line.lstrip())
                if comment_line and not comment_line.startswith(cs):
                    comment_line = csp + comment_line
                out.append(indent_string + comment_line)
            this_entry = section[entry]
            comment = self._handle_comment(section.inline_comments[entry])

            if isinstance(this_entry, Section):
                # a section
                out.append(self._write_marker(
                    indent_string,
                    this_entry.depth,
                    entry,
                    comment))
                out.extend(self.write(section=this_entry))
            else:
                out.append(self._write_line(
                    indent_string,
                    entry,
                    this_entry,
                    comment))
        if self.is_new_file:
            out.append("")

        if section is self:
            for line in self.final_comment:
                line = self._decode_element(line)
                stripped_line = line.strip()
                if stripped_line and not stripped_line.startswith(cs):
                    line = csp + line
                out.append(line)
            self.interpolation = int_val

        if section is not self:
            return out

        if (self.filename is None) and (outfile is None):
            # output a list of lines
            # might need to encode
            # NOTE: This will *screw* UTF16, each line will start with the BOM
            if self.encoding:
                out = [l.encode(self.encoding) for l in out]
            if (self.BOM and ((self.encoding is None) or
                              (BOM_LIST.get(self.encoding.lower()) == 'utf_8'))):
                # Add the UTF8 BOM
                if not out:
                    out.append('')
                out[0] = BOM_UTF8 + out[0]
            return out

        # Turn the list to a string, joined with correct newlines
        newline = self.newlines or os.linesep
        if (getattr(outfile, 'mode', None) is not None and outfile.mode == 'w'
                and sys.platform == 'win32' and newline == '\r\n'):
            # Windows specific hack to avoid writing '\r\r\n'
            newline = '\n'
        output = self._a_to_u(newline).join(out)
        if not output.endswith(newline):
            output += newline

        if isinstance(output, six.binary_type):
            output_bytes = output
        else:
            output_bytes = output.encode(self.encoding or
                                         self.default_encoding or
                                         'ascii')

        if self.BOM and ((self.encoding is None) or match_utf8(self.encoding)):
            # Add the UTF8 BOM
            output_bytes = BOM_UTF8 + output_bytes

        if outfile is not None:
            outfile.write(output_bytes)
        else:
            with open(self.filename, 'wb') as h:
                h.write(output_bytes)

    def dirname(self):
        return os.path.dirname(os.path.abspath(self.filename))

    def update_attribute(self, attribute_name, value, section_name=None):
        if section_name:
            if section_name not in self:
                self[section_name] = dict()
            self[section_name][attribute_name] = value
        else:
            self[attribute_name] = value

    def _parse(self, infile):
        """Actually parse the config file."""
        temp_list_values = self.list_values
        if self.unrepr:
            self.list_values = False

        comment_list = []
        done_start = False
        this_section = self
        maxline = len(infile) - 1
        cur_index = -1
        reset_comment = False

        while cur_index < maxline:
            if reset_comment:
                comment_list = []
            cur_index += 1
            line = infile[cur_index]
            sline = line.strip()
            # do we have anything on the line ?
            if not sline or sline.startswith(('#', ';', '/')):
                reset_comment = False
                comment_list.append(line)
                continue

            if not done_start:
                # preserve initial comment
                self.initial_comment = comment_list
                comment_list = []
                done_start = True

            reset_comment = True
            # first we check if it's a section marker
            mat = self._sectionmarker.match(line)
            if mat is not None:
                # is a section line
                (indent, sect_open, sect_name, sect_close, comment) = mat.groups()
                if indent and (self.indent_type is None):
                    self.indent_type = indent
                cur_depth = sect_open.count('[')
                if cur_depth != sect_close.count(']'):
                    self._handle_error("Cannot compute the section depth",
                                       configobj.NestingError, infile, cur_index)
                    continue

                if cur_depth < this_section.depth:
                    # the new section is dropping back to a previous level
                    try:
                        parent = self._match_depth(this_section,
                                                   cur_depth).parent
                    except SyntaxError:
                        self._handle_error("Cannot compute nesting level",
                                           configobj.NestingError, infile, cur_index)
                        continue
                elif cur_depth == this_section.depth:
                    # the new section is a sibling of the current section
                    parent = this_section.parent
                elif cur_depth == this_section.depth + 1:
                    # the new section is a child the current section
                    parent = this_section
                else:
                    self._handle_error("Section too nested",
                                       configobj.NestingError, infile, cur_index)
                    continue

                sect_name = self._unquote(sect_name)
                if sect_name in parent:
                    self._handle_error('Duplicate section name',
                                       configobj.DuplicateError, infile, cur_index)
                    continue

                # create the new section
                this_section = Section(
                    parent,
                    cur_depth,
                    self,
                    name=sect_name)
                parent[sect_name] = this_section
                parent.inline_comments[sect_name] = comment
                parent.comments[sect_name] = comment_list
                continue
            #
            # it's not a section marker,
            # so it should be a valid ``key = value`` line
            mat = self._keyword.match(line)
            if mat is None:
                self._handle_error(
                    'Invalid line ({0!r}) (matched as neither section nor keyword)'.format(line),
                    configobj.ParseError, infile, cur_index)
            else:
                # is a keyword value
                # value will include any inline comment
                (indent, key, value) = mat.groups()
                if indent and (self.indent_type is None):
                    self.indent_type = indent
                # check for a multiline value
                if value[:3] in ['"""', "'''"]:
                    try:
                        value, comment, cur_index = self._multiline(
                            value, infile, cur_index, maxline)
                    except SyntaxError:
                        self._handle_error(
                            'Parse error in multiline value',
                            configobj.ParseError, infile, cur_index)
                        continue
                    else:
                        if self.unrepr:
                            comment = ''
                            try:
                                value = unrepr(value)
                            except Exception as e:
                                if type(e) == configobj.UnknownType:
                                    msg = 'Unknown name or type in value'
                                else:
                                    msg = 'Parse error from unrepr-ing multiline value'
                                self._handle_error(msg, configobj.UnreprError, infile,
                                                   cur_index)
                                continue
                else:
                    if self.unrepr:
                        comment = ''
                        try:
                            value = unrepr(value)
                        except Exception as e:
                            if isinstance(e, configobj.UnknownType):
                                msg = 'Unknown name or type in value'
                            else:
                                msg = 'Parse error from unrepr-ing value'
                            self._handle_error(msg, configobj.UnreprError, infile,
                                               cur_index)
                            continue
                    else:
                        # extract comment and lists
                        try:
                            (value, comment) = self._handle_value(value)
                        except SyntaxError:
                            self._handle_error(
                                'Parse error in value',
                                configobj.ParseError, infile, cur_index)
                            continue
                #
                key = self._unquote(key)
                if key in this_section:
                    self._handle_error(
                        'Duplicate keyword name',
                        configobj.DuplicateError, infile, cur_index)
                    continue
                # add the key.
                # we set unrepr because if we have got this far we will never
                # be creating a new section
                this_section.__setitem__(key, value, unrepr=True)
                this_section.inline_comments[key] = comment
                this_section.comments[key] = comment_list
                continue
        #
        if self.indent_type is None:
            # no indentation used, set the type accordingly
            self.indent_type = ''

        # preserve the final comment
        if not self and not self.initial_comment:
            self.initial_comment = comment_list
        elif not reset_comment:
            self.final_comment = comment_list
        self.list_values = temp_list_values


def extract_ini_primitive_value(returned_val, cast_type=None):
    if isinstance(returned_val, list):
        returned_val = returned_val[0]
    if cast_type:
        return cast_type(returned_val.split()[0])
    return returned_val.split()[0]



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import math
import struct
import sqlite3
import toml
import logging

from collections import OrderedDict
import sim_racing_tools.automation.installation as installation

SAMPLE_VARIANT_UID = 'B70604DD4EF0BE1E016E1F9559D67659'
SANDBOX_DB_FILE_PATH = None

schema = dict()


class SQLiteColumn(object):
    def __init__(self):
        self.name = None
        self.sqlite_type = None

    def __str__(self):
        return f'{self.sqlite_type} called {self.name}'


def load_engine_results_schema():
    with sqlite3.connect(installation.get_sandbox_db_path()) as conn:
        cur = conn.cursor()
        for row in cur.execute('PRAGMA table_info("EngineResults")').fetchall():
            a = SQLiteColumn()
            a.name = row[1]
            a.sqlite_type = row[2]
            schema[a.name] = a


def get_resource_data(variant_uid):
    resource_stats_to_collect = ["EngineeringCost", "EngineeringTime", "ManHours",
                                 "MaterialCost", "ToolingCosts", "TotalCost"]
    data_dict = OrderedDict()
    with sqlite3.connect(installation.get_sandbox_db_path()) as conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        row = cur.execute('SELECT * FROM EngineResults WHERE UID = ?', (variant_uid,)).fetchone()
        for stat in resource_stats_to_collect:
            data_dict[stat] = row[stat]
    return data_dict


def write_resource_summary(variant_uid, out_file):
    with open(out_file, "w+") as f:
        toml.dump(get_resource_data(variant_uid), f)


def _decode_double(blob_bytes):
    logging.debug(" ".join(f'{b:02X}' for b in blob_bytes))
    return struct.unpack("d", bytearray(blob_bytes))[0]


def _decode_blob(blob_bytes):
    # skip blob header
    logging.debug(f'Blob length: {len(blob_bytes)}')
    data = blob_bytes[2:]
    num_data_points = struct.unpack("<LL", bytearray(data[0:8]))[0]
    logging.debug(f'Num data points: {num_data_points}')
    data = data[8:]
    current_pos = 0
    collected_points = 0
    out_list = list()
    while collected_points < num_data_points:
        logging.debug(f'{data[current_pos]:02X}')
        current_pos += 1
        logging.debug(f'{_decode_double(data[current_pos:current_pos+8])}')
        current_pos += 9
        out_list.append(_decode_double(data[current_pos:current_pos+8]))
        current_pos += 8
        collected_points += 1
    return out_list


def get_engine_graph_data(variant_uid):
    with sqlite3.connect(installation.get_sandbox_db_path()) as conn:
        conn.row_factory = sqlite3.Row
        conn.text_factory = bytes
        cur = conn.cursor()
        data = cur.execute('SELECT * from EngineCurves where uid = ?', (variant_uid,)).fetchone()
        collection_data = {"rpm": 'RPMCurve',
                           "power": 'PowerCurve',
                           "torque": 'TorqueCurve',
                           'econ': 'EconCurve',
                           'econ-eff': 'EconEffCurve',
                           'boost': 'BoostCurve'}
        data_dict = OrderedDict()
        for header, column_name in collection_data.items():
            logging.debug(f'{header} data')
            logging.debug(" ".join(f'{b:02X}' for b in data[column_name]))
            data_dict[header] = _decode_blob(data[column_name])
        return data_dict


def get_trim_graph_data(trim_uid):
    with sqlite3.connect(installation.get_sandbox_db_path()) as conn:
        conn.row_factory = sqlite3.Row
        conn.text_factory = bytes
        cur = conn.cursor()
        data = cur.execute('SELECT * from TrimGraphData where uid = ?', (trim_uid,)).fetchone()
        graph_data = _decode_blob(data["GraphData"])
    return graph_data


def write_engine_performance_summary(variant_uid, out_file):
    engine_graph_data = get_engine_graph_data(variant_uid)
    with open(out_file, "w+") as f:
        toml.dump(engine_graph_data, f)
    with open("power.lut", "w+") as f:
        for idx in range(0, len(engine_graph_data["rpm"])):
            f.write(f'{engine_graph_data["rpm"][idx]}|{math.ceil(engine_graph_data["torque"][idx])}\n')


if __name__ == '__main__':
    write_resource_summary(sys.argv[1], "resource_summary.toml")
    write_engine_performance_summary(sys.argv[1], "performance_summary.toml")



import sys
import sim_racing_tools.assetto_corsa.car as car
from sim_racing_tools.assetto_corsa.car.drivetrain import EfficiencyLookup
from sim_racing_tools.automation.fabricator.assetto_corsa import DefaultEngineFabricator

SUCCESS = 0
FAIL = 1


def clone_car(args):
    try:
        car.create_new_car_from_existing(args.new_brand_name, args.new_model_name, args.ac_car_folder)
        return SUCCESS
    except Exception as e:
        print(f"Failed to clone {args.ac_car_folder}")
        print(str(e), file=sys.stderr)
        return FAIL


def swap_engine(args):
    try:
        ac_car = car.load_car(args.ac_car_folder)
        fabricator = DefaultEngineFabricator(use_csp_physics_extensions=args.use_csp_physics,
                                             mechanical_efficiency=EfficiencyLookup[ac_car.drivetrain.drive_type])
        ac_car.swap_engine(fabricator.create_from_beamng_mod(args.exported_car_folder),
                           update_mass=args.update_mass,
                           old_engine_mass=args.mass_hint,
                           use_csp_physics_extensions=args.use_csp_physics)
        ac_car.write()
    except Exception as e:
        print(f"Failed to swap engine from {args.exported_car_folder} to {args.ac_car_folder}")
        import traceback
        traceback.print_exc()
        print(str(e), file=sys.stderr)
        return FAIL



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import logging

LIST = 0
DICT = 1


#  Start by looking for key starting with "
#  Once found there will then be a ":" then it could be could be one of 4 things:
#    string - "
#    dict - {
#    list - [
#    number - any other char
class ParseState(object):
    ATTRIBUTE_NAME_SEARCH = 0
    ATTRIBUTE_VALUE_SEARCH = 1


class NewLineFound(IndexError):
    def __init__(self):
        super(NewLineFound, self).__init__("Found new line")


class Parser(object):
    def __init__(self):
        self.data_dict = dict()
        self.container_stack = [self.data_dict]
        self.current_container = self.container_stack[-1]
        self.line = None
        self.idx = 0
        self.current_attribute_name = None

        self.state = ParseState.ATTRIBUTE_NAME_SEARCH

    def naive_parse(self, filename):
        with open(filename, "r") as f:
            return self._naive_parse(f.readlines(), logging.getLogger("jbeam_load"))

    def _naive_parse(self, jbeam_lines, logger):
        line_idx = 0
        try:
            for line_idx, line in enumerate(jbeam_lines, start=1):
                self.idx = 0
                if line[0] == "{":
                    continue
                self.line = line.strip()
                comment_start = self.line.find("//")
                if comment_start >= 0:
                    comment = line[comment_start:-1]
                    logger.info(f"Removing {comment} from line {line_idx}")
                    self.line = self.line[:comment_start] + "\n"
                self._parse_line()
        except Exception as e:
            err_msg = f"Error on line {line_idx}: {str(e)}\n {self.line}"
            logger.error(err_msg)
            raise ValueError(err_msg)
        return self.data_dict

    def _parse_string(self):
        if self.line[self.idx] != '"':
            raise ValueError("String doesn't start with a \"")
        self.idx += 1
        ending_quote_idx = self.line[self.idx:].index('"')
        string = self.line[self.idx:self.idx+ending_quote_idx]
        self.idx += ending_quote_idx + 1
        return string

    def _parse_number(self):
        is_float = False
        found_non_digit = False
        first_char = True
        for number_end, char in enumerate(self.line[self.idx:]):
            if first_char:
                first_char = False
                if char == "-":
                    continue
            if char.isdigit():
                continue
            if char == ".":
                if is_float:
                    raise ValueError("Unknown number type containing multiple decimal places")
                is_float = True
            else:
                found_non_digit = True
                break
        if not found_non_digit:
            number_end += 1
        if is_float:
            val = float(self.line[self.idx:self.idx+number_end])
        else:
            val = int(self.line[self.idx:self.idx+number_end])
        self.idx += number_end
        return val

    def _parse_bool(self):
        if self.line[self.idx] in ["T", "t"]:
            val = bool(self.line[self.idx:self.idx+4])
            self.idx += 4
            return val
        else:
            val = bool(self.line[self.idx:self.idx+5])
            self.idx += 5
            return val

    def _found_container_close(self):
        return self.line[self.idx] in ["}", "]"]

    def _parse_container_close(self):
        self._pop_container()
        self.idx += 1

    def _parse_line(self):
        while self.idx < len(self.line):
            if self.state == ParseState.ATTRIBUTE_NAME_SEARCH:
                try:
                    self.idx = self._find_next_char()
                except (NewLineFound, ValueError):
                    return

                if self._found_container_close():
                    self._parse_container_close()
                    continue

                if self.line[self.idx] != '"':
                    raise ValueError(f"Was expecting an attribute name but found {self.line[self.idx]} at position "
                                     f"{self.idx}")
                attribute_name = self._parse_string()
                self.idx = self.idx + self.line[self.idx:].index(":") + 1
                self._attribute_name_found(attribute_name)
                continue

            try:
                self.idx = self._find_next_char()
            except (NewLineFound, ValueError):
                return

            if self.line[self.idx] == '"':
                self._add_value(self._parse_string())
                continue
            elif self.line[self.idx].isdigit() or self.line[self.idx] == "-":
                self._add_value(self._parse_number())
                continue
            elif self.line[self.idx] in ["T", "t", "F", "f"]:
                self._add_value(self._parse_bool())
                continue
            elif self._found_container_close():
                self._parse_container_close()
                continue

            if self.line[self.idx] == '{':
                self._add_container(dict())
            elif self.line[self.idx] == '[':
                self._add_container(list())
            else:
                raise ValueError(f"Unexpected char {self.line[self.idx]} at position {self.idx}")
            self.idx += 1

    def _attribute_name_found(self, name):
        self.state = ParseState.ATTRIBUTE_VALUE_SEARCH
        self.current_attribute_name = name

    def _attribute_stored(self):
        self.state = ParseState.ATTRIBUTE_NAME_SEARCH
        self.current_attribute_name = None

    def _current_container_type(self):
        return type(self.current_container)

    def _find_next_char(self):
        for idx, char in enumerate(self.line[self.idx:]):
            if char == "\n":
                raise NewLineFound()
            if char.isspace() or char == ",":
                continue
            return self.idx + idx
        raise ValueError("Couldn't find a non-space char on current line")

    def _push_container(self, container):
        self.container_stack.append(container)
        self.current_container = self.container_stack[-1]
        self._set_state_for_current_container()

    def _pop_container(self):
        self.container_stack.pop()
        if len(self.container_stack):
            self.current_container = self.container_stack[-1]
        self._set_state_for_current_container()

    def _add_container(self, container):
        if self._current_container_type() == list:
            self.current_container.append(container)
            self._push_container(self.current_container[-1])
        else:
            if self.current_attribute_name is None:
                raise ValueError(f"Cannot add {type(container)} to dict without a key")
            self.current_container[self.current_attribute_name] = container
            self._push_container(self.current_container[self.current_attribute_name])
            self.current_attribute_name = None

    def _add_value(self, value):
        if self._current_container_type() == list:
            self.current_container.append(value)
        else:
            if self.current_attribute_name is None:
                raise ValueError(f"Cannot add {type(value)} to dict without a key")
            self.current_container[self.current_attribute_name] = value
            self.state = ParseState.ATTRIBUTE_NAME_SEARCH
            self.current_attribute_name = None

    def _set_state_for_current_container(self):
        if self._current_container_type() == dict:
            self.state = ParseState.ATTRIBUTE_NAME_SEARCH
        else:
            self.state = ParseState.ATTRIBUTE_VALUE_SEARCH



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""
import os
import re
import zipfile
import decimal
import locale
import struct


def create_filename_safe_name(in_name):
    """
    Take a string and make it safe to use in a file path by removing any non-letter and number characters and
    replacing any spaces with single underscores

    Args:
        in_name: the name to adapt

    Returns:
        a string that is suitable for using in file paths
    """
    normalized = re.sub(r"[^\w\s]", '', in_name)
    return re.sub(r"\s+", '_', normalized)


def round_up(x, place=0):
    context = decimal.getcontext()
    # get the original setting so we can put it back when we're done
    original_rounding = context.rounding
    # change context to act like ceil()
    context.rounding = decimal.ROUND_CEILING

    rounded = round(decimal.Decimal(str(x)), place)
    context.rounding = original_rounding
    return float(rounded)


def kw_to_bhp(kw):
    return kw / 0.745699872


def unzip_file(zip_file_path):
    if os.path.isfile(zip_file_path):
        with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
            zip_ref.extractall(os.path.join(os.path.dirname(zip_file_path),
                                            os.path.basename(zip_file_path)[0:-4]))
    else:
        raise IOError(f"No such file: {zip_file_path}")


# https://gist.github.com/Winand/997ed38269e899eb561991a0c663fa49
def read_win_shortcut(path):
    # http://stackoverflow.com/a/28952464/1119602
    with open(path, 'rb') as stream:
        content = stream.read()
        # skip first 20 bytes (HeaderSize and LinkCLSID)
        # read the LinkFlags structure (4 bytes)
        lflags = struct.unpack('I', content[0x14:0x18])[0]
        position = 0x18
        # if the HasLinkTargetIDList bit is set then skip the stored IDList
        # structure and header
        if (lflags & 0x01) == 1:
            position = struct.unpack('H', content[0x4C:0x4E])[0] + 0x4E
        last_pos = position
        position += 0x04
        # get how long the file information is (LinkInfoSize)
        length = struct.unpack('I', content[last_pos:position])[0]
        # skip 12 bytes (LinkInfoHeaderSize, LinkInfoFlags and VolumeIDOffset)
        position += 0x0C
        # go to the LocalBasePath position
        lbpos = struct.unpack('I', content[position:position + 0x04])[0]
        position = last_pos + lbpos
        # read the string at the given position of the determined length
        size = (length + last_pos) - position - 0x02
        content = content[position:position + size].split(b'\x00', 1)
        return content[-1].decode('utf-16' if len(content) > 1
                                  else locale.getdefaultlocale()[1])



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
import argcomplete
import argparse

parser = argparse.ArgumentParser(description='Tools for working with Assetto Corsa (AC) content')

subparsers = parser.add_subparsers(title='Commands')
parser_clone_car = subparsers.add_parser('clone-car',
                                         help="Clone a car from the AC cars directory. Note: You must have already "
                                              "extracted the data.acd file to a data directory")
parser_clone_car.add_argument("ac_car_folder",
                              help="The name of the folder inside the AC cars directory to clone")
parser_clone_car.add_argument("new_brand_name", help="The name of the brand the newly cloned car will have")
parser_clone_car.add_argument("new_model_name", help="Then name of the model the newly cloned car will have")

parser_swap_engine = subparsers.add_parser('swap-automation-engine',
                                           help="Takes an engine from an automation car exported to BeamNG, creates "
                                                "the AC engine parameter data and puts that data into the provided car "
                                                "folder")
parser_swap_engine.add_argument('ac_car_folder', help="The AC car to swap the engine into")
parser_swap_engine.add_argument('exported_car_folder',
                                help="The name of the folder in the BeamNG mod directory containing the BeamNG "
                                     "exported car with the engine")
parser_swap_engine.add_argument('-m', "--update-mass", action="store_true",
                                help="Will update the car mass based on the difference between the old "
                                     "and new engine")
parser_swap_engine.add_argument("--mass-hint", type=int,
                                help="If the engine in the car has no mass data (i.e. isn't an engine "
                                     "created from Automation data) then you have to provide a mass "
                                     "for the old engine if you want to update the car mass. The value "
                                     "of this parameter will be used as the old engine mass in this "
                                     "scenario - the values unit is Kg")
parser_swap_engine.add_argument('-c', '--use-csp-physics', action="store_true",
                                help="Allow use of custom shader patch physics extensions when creating engine data")
argcomplete.autocomplete(parser)


def main():
    import sys
    import sim_racing_tools.assetto_corsa.scripts.ac_tools_impl as ac_impl
    parser_clone_car.set_defaults(func=ac_impl.clone_car)
    parser_swap_engine.set_defaults(func=ac_impl.swap_engine)
    args = parser.parse_args()
    sys.exit(args.func(args))


if __name__ == '__main__':
    main()



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import os
import struct
import sqlite3
import logging
from collections import OrderedDict
import sim_racing_tools.automation.installation as installation


def get_engine_bill_of_material_params():
    return {"EngineeringCost", "EngineeringTime", "ManHours",
            "MaterialCost", "ToolingCosts", "TotalCost", "ServiceCost"}


def get_engine_bill_of_materials(variant_uid):
    resource_stats_to_collect = get_engine_bill_of_material_params()
    data_dict = OrderedDict()
    with sqlite3.connect(os.path.join(installation.get_userdata_path(),
                                      installation.SANDBOX_DB_NAME)) as conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        row = cur.execute('SELECT * FROM EngineResults WHERE UID = ?', (variant_uid,)).fetchone()
        for stat in resource_stats_to_collect:
            data_dict[stat] = row[stat]
    return data_dict


def get_engine_graph_data_params():
    return {'RPMCurve', 'PowerCurve', 'TorqueCurve', 'EconCurve', 'EconEffCurve', 'BoostCurve'}


def get_engine_graph_data(variant_uid):
    with sqlite3.connect(os.path.join(installation.get_userdata_path(),
                                      installation.SANDBOX_DB_NAME)) as conn:
        conn.row_factory = sqlite3.Row
        conn.text_factory = bytes
        cur = conn.cursor()
        data = cur.execute('SELECT * from EngineCurves where uid = ?', (variant_uid,)).fetchone()
        data_dict = OrderedDict()
        for header in get_engine_graph_data_params():
            # logging.debug(f'{header} data')
            # logging.debug(" ".join(f'{b:02X}' for b in data[column_name]))
            data_dict[header] = _decode_blob(data[header])
        return data_dict


def get_engine_performance_data_params():
    return {"AverageCruiseEcon", "Econ", "MinEcon", "WorstEcon", "EconEff", "IdleSpeed", "MaxRPM",
            "MTTF", "Weight", "PeakTorque", "PeakTorqueRPM", "PeakPower", "PeakPowerRPM",
            "PeakBoost", "PeakBoostRPM"}


def get_engine_performance_data(variant_uid):
    with sqlite3.connect(os.path.join(installation.get_userdata_path(),
                                      installation.SANDBOX_DB_NAME)) as conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        engine_result_columns = get_engine_performance_data_params()
        query = f"SELECT {', '.join(engine_result_columns)} FROM EngineResults WHERE uid = ?"
        return cur.execute(query, (variant_uid,)).fetchone()


def get_variant_data_params():
    return {"UID", "FUID", "Name", "GameVersion", "Crank", "Conrods", "Pistons", "VVT", "Aspiration",
            "AspirationType", "AspirationOption", "IntercoolerSetting", "FuelSystemType", "FuelSystem",
            "IntakeManifold", "Intake", "FuelType", "Headers", "ExhaustCount", "ExhaustBypassValves",
            "Cat", "Muffler1", "Muffler2", "Bore", "Stroke", "Capacity", "Compression",
            "CamProfileSetting", "VVLCamProfileSetting", "AFR", "AFRLean", "RPMLimit",
            "IgnitionTimingSetting", "ARRatio", "BoostCutOff", "CompressorFraction", "TurbineFraction",
            "ExhaustDiameter", "QualityBottomEnd", "QualityTopEnd", "QualityAspiration",
            "QualityFuelSystem", "QualityExhaust", "Tags"}


def get_family_mapping():
    return {"Name": "FamilyName", "BlockConfig": "BlockConfig", "BlockMaterial": "BlockMaterial",
            "BlockType": "BlockType", "Head": "Head", "HeadMaterial": "HeadMaterial", "Valves": "Valves",
            "VVL": "VVL"}


def get_family_data_params():
    return set(get_family_mapping().values())


def get_engine_data_params():
    params = get_variant_data_params()
    params.update(get_engine_performance_data_params())
    params.update(get_engine_bill_of_material_params())
    params.update(get_engine_graph_data_params())
    params.update(get_family_data_params())
    return params


def get_engine_data(variant_uid):
    with sqlite3.connect(os.path.join(installation.get_userdata_path(),
                                      installation.SANDBOX_DB_NAME)) as conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        variant_columns = get_variant_data_params()
        query = f"SELECT {', '.join(variant_columns)} from Variants where uid = ?"
        row = cur.execute(query, (variant_uid,)).fetchone()
        data_dict = {k: row[k] for k in row.keys()}
        family_row = cur.execute('SELECT * from Families where uid = ?',
                                 (data_dict['FUID'],)).fetchone()
        for family_key, data_key in get_family_mapping().items():
            data_dict[data_key] = family_row[family_key]
        data_dict.update(get_engine_performance_data(variant_uid))
        data_dict.update(get_engine_bill_of_materials(variant_uid))
        data_dict.update(get_engine_graph_data(variant_uid))
    return data_dict


def get_engine_uid_from_name(family_name, variant_name):
    with sqlite3.connect(os.path.join(installation.get_userdata_path(),
                                      installation.SANDBOX_DB_NAME)) as conn:
        conn.row_factory = sqlite3.Row
        cur = conn.cursor()
        family_row = cur.execute('SELECT * from Families where Name = ?',
                                 (family_name,)).fetchone()
        query = f"SELECT UID from Variants where FUID = ? and Name = ?"
        return cur.execute(query, (family_row['UID'], variant_name)).fetchone()["UID"]


def get_engine_by_name(family_name, variant_name):
    return get_engine_data(get_engine_uid_from_name(family_name, variant_name))


def _decode_double(blob_bytes):
    logging.debug(" ".join(f'{b:02X}' for b in blob_bytes))
    return struct.unpack("d", bytearray(blob_bytes))[0]


def _decode_blob(blob_bytes):
    # skip blob header
    logging.debug(f'Blob length: {len(blob_bytes)}')
    data = blob_bytes[2:]
    num_data_points = struct.unpack("<LL", bytearray(data[0:8]))[0]
    logging.debug(f'Num data points: {num_data_points}')
    data = data[8:]
    current_pos = 0
    collected_points = 0
    out_list = list()
    while collected_points < num_data_points:
        logging.debug(f'{data[current_pos]:02X}')
        current_pos += 1
        logging.debug(f'{_decode_double(data[current_pos:current_pos+8])}')
        current_pos += 9
        out_list.append(_decode_double(data[current_pos:current_pos+8]))
        current_pos += 8
        collected_points += 1
    return out_list



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
import shutil
import json

import sim_racing_tools.utils as utils
import sim_racing_tools.assetto_corsa.utils as ac_utils
from sim_racing_tools.assetto_corsa.utils import extract_ini_primitive_value
import sim_racing_tools.assetto_corsa.installation as installation
import sim_racing_tools.assetto_corsa.car.engine as engine
import sim_racing_tools.assetto_corsa.car.drivetrain as drivetrain
import sim_racing_tools.quick_bms as quick_bms
from typing import List


CSP_EXTENDED_PHYSICS_VERSION = "extended-2"


class NoSuchCar(ValueError):
    def __init__(self, car_name):
        super(NoSuchCar, self).__init__(f"No car exists with the name {car_name}")


class CannotUpdateMass(ValueError):
    def __init__(self):
        super(CannotUpdateMass, self).__init__("Can't update car weight as no engine weight is available")


def load_car(car_folder_name):
    ac_install = installation.Installation()
    if car_folder_name not in ac_install.installed_cars:
        raise NoSuchCar(car_folder_name)
    c = Car()
    c.load_from_path(os.path.join(ac_install.get_installed_cars_path(), car_folder_name))
    return c


def create_new_car_from_existing(brand, model, existing_car_folder_name, folder_prefix=None, folder_name=None):
    ac_install = installation.Installation()

    if not folder_name:
        folder_name = _create_car_folder_name(brand, model, folder_prefix)
    folder_name = utils.create_filename_safe_name(folder_name)
    logging.basicConfig(filename=f'create_{folder_name}.log', level=logging.INFO)
    logging.getLogger()

    new_car_path = _create_copy_of_existing_car_directory(ac_install, existing_car_folder_name, folder_name)
    if not os.path.isdir(os.path.join(new_car_path, installation.DATA_FOLDER_NAME)):
        # todo find and run openBMS (on wine if linux)
        logging.error(f"The existing car does not have an extracted data folder. "
                      f"The {installation.ENCODED_DATA_FILENAME} needs to be extracted with "
                      f"{quick_bms.QUICKBMS_EXE_NAME}")
        raise NotImplementedError("TODO run openBMS on data.acd")
    acd_file_path = os.path.join(new_car_path, installation.ENCODED_DATA_FILENAME)
    if os.path.exists(acd_file_path):
        logging.info(f"Removing {acd_file_path}")
        os.remove(acd_file_path)

    _update_filenames_and_references(new_car_path, existing_car_folder_name, brand, model)
    _update_car_sfx_dir(ac_install, new_car_path, existing_car_folder_name)


def _create_car_folder_name(brand: str, model_name:str, folder_prefix=None):
    folder_name = "" if not folder_prefix else f'{folder_prefix}_'
    folder_name += f'{brand}_{model_name}'
    return folder_name.lower()


def _create_copy_of_existing_car_directory(ac_install: installation.Installation,
                                           existing_car_path: str,
                                           new_car_path: str):
    if existing_car_path not in ac_install.installed_cars:
        err_str = f"{existing_car_path} is not present in {ac_install.get_installed_cars_path()}"
        logging.error(err_str)
        raise ValueError(err_str)
    existing_car_full_path = os.path.join(ac_install.get_installed_cars_path(), existing_car_path)
    new_car_full_path = os.path.join(ac_install.get_installed_cars_path(), new_car_path)
    if os.path.exists(new_car_full_path):
        err_str = f"There is already a car of this name present at {new_car_full_path}"
        logging.error(err_str)
        raise ValueError(err_str)
    logging.info(f'Copying contents of {existing_car_full_path} to {new_car_full_path}')
    shutil.copytree(existing_car_full_path, new_car_full_path)
    return new_car_full_path


def _update_filenames_and_references(new_car_path: str,
                                     old_name: str,
                                     brand: str,
                                     model: str):
    new_car_dir_name = os.path.basename(new_car_path)
    for dirname, dirs, files in os.walk(new_car_path):
        for filename in files:
            full_file_path = os.path.join(dirname, filename)
            if filename.startswith(old_name) and (filename.endswith(".kn5") or filename.endswith(".bank")):
                new_file_path = os.path.join(dirname, filename.replace(old_name, new_car_dir_name))
                logging.info(f"Renaming {full_file_path} to {new_file_path}")
                os.rename(full_file_path, new_file_path)
            elif filename == "car.ini":
                config = ac_utils.IniObj(full_file_path)
                new_screen_name = f'{brand} {model}'
                logging.info(f"Updating SCREEN_NAME in car.ini to {new_screen_name}")
                config["INFO"]["SCREEN_NAME"] = new_screen_name
                config.write()
            elif filename == "lods.ini":
                config = ac_utils.IniObj(full_file_path)
                idx = 0
                while True:
                    current_lod_name = f'LOD_{idx}'
                    if current_lod_name not in config:
                        break
                    lod_section = config[current_lod_name]
                    if "FILE" in lod_section:
                        tmp = lod_section["FILE"]
                        logging.info(f"Updating FILE in {current_lod_name} section of {full_file_path} from "
                                     f"{old_name} to {new_car_dir_name}")
                        lod_section["FILE"] = tmp.replace(old_name, new_car_dir_name)
                    idx += 1
                config.write()
            elif filename == "ui_car.json":
                with open(full_file_path, "r") as ui_car_file:
                    config = json.load(ui_car_file, strict=False)
                    logging.info("Updating ui_car.json")
                    new_name = f'{brand} {model}'
                    logging.info(f"Updating {config['name']} to {new_name}")
                    config["name"] = new_name
                    logging.info(f"Updating {config['brand']} to {brand}")
                    config["brand"] = brand
                with open(full_file_path, "w+") as ui_car_file:
                    ui_car_file.write(json.dumps(config, indent=4))


def _update_car_sfx_dir(ac_install: installation.Installation,
                        new_car_path: str,
                        existing_car_folder_name: str):
    new_car_dirname = os.path.basename(new_car_path)
    guids_file_path = os.path.join(new_car_path, os.sep.join(["sfx", installation.SFX_GUID_FILENAME]))
    if os.path.isfile(guids_file_path):
        logging.info(f"Existing {installation.SFX_GUID_FILENAME} file present")
        with open(guids_file_path, "r") as f:
            existing_guid_lines = f.readlines()

        def _update_guid_line(guid_line):
            updated_line = guid_line.replace(existing_car_folder_name, new_car_dirname)
            logging.info(f"Updating {guid_line} to {updated_line}")
            return updated_line

        with open(guids_file_path, "w+") as f:
            f.writelines([_update_guid_line(line) for line in existing_guid_lines])
    else:
        logging.info(f"No {installation.SFX_GUID_FILENAME} file present. A new one will be created")
        with open(guids_file_path, "w+") as guid_file:
            try:
                guid_file.writelines(_generate_guids_content(ac_install, new_car_dirname, existing_car_folder_name))
            except IndexError:
                logging.warning(f"Can't find any existing sfx data for {existing_car_folder_name} - new car won't "
                                f"have any sfx")


def _generate_guids_content(ac_install: installation.Installation,
                            new_car_dir_name: str,
                            old_car_dir_name: str):
    sfx_list = list()
    sfx_list.append(f'{ac_install.sfx_bank_dict[old_car_dir_name]} bank:/{new_car_dir_name}\n')
    for sfx_entry in ac_install.sfx_dict_by_folder_name[old_car_dir_name]:
        sfx_list.append(sfx_entry.replace(old_car_dir_name, new_car_dir_name))
    return sfx_list


class Car(object):
    def __init__(self):
        self.car_path = None
        self.data_path = None
        self.car_ini_data = None
        self.ui_info = UIInfo()
        self.version: str = '1'
        self.screen_name: str = ""
        self.total_mass: int = 0  # total vehicle weight in kg with driver and no fuel
        '''
        sql EngineResults.Econ = BSFC (g/(kWh))
        EconEff = 1 / (BSFC * "LHV of fuel")
        LHC of gasoline from wikipedia -> 0.0122222
        bsfc = (fuel consumption rate) / (Power output) = M.f / Pwrb kg/kWh
        1680.8221 = r / 167236.78
        Q = N * q / R, where
        Q – (in l / h) maximum theoretical fuel consumption in grams per 1 hour of engine operation at maximum power;
        q – (in g / kWh) specific fuel consumption for power N;
        N – (in kW) engine power;
        R – (kg / m3) fuel density
        Gasoline (petrol) fuel density: 710 – 760 kg / m3
        fuel consumption. In one second the consumption is (rpm*gas*CONSUMPTION)/1000 litres
        is gas 0 - 1 or 0 - 100?
        0.175698056 = (8600 * 100 * C) / 100
        632.513 l/h
        '''
        self.fuel_consumption: float = 0.0
        self.default_fuel: int = 0  # default starting fuel in litres
        self.max_fuel: int = 0  # max fuel in litres
        self.ai_shift_up: int = 0
        self.ai_shift_down: int = 0
        self.engine: engine.Engine or None = None
        self.drivetrain: drivetrain.Drivetrain or None = None
        self.shift_lights: ShiftLights = ShiftLights()

    @property
    def using_csp_extended_physics(self):
        return self.version == CSP_EXTENDED_PHYSICS_VERSION

    @using_csp_extended_physics.setter
    def using_csp_extended_physics(self, use):
        if use:
            self.version = CSP_EXTENDED_PHYSICS_VERSION
        else:
            self.version = "1"

    def load_from_path(self, car_path):
        self.car_path = car_path
        self.data_path = os.path.join(car_path, "data")
        ini_data = ac_utils.IniObj(os.path.join(self.data_path, "car.ini"))
        self.version = ini_data["HEADER"]["VERSION"]
        self.screen_name = ini_data["INFO"]["SCREEN_NAME"]
        self.total_mass = extract_ini_primitive_value(ini_data["BASIC"]["TOTALMASS"], int)
        if "CONSUMPTION" in ini_data["FUEL"]:
            self.fuel_consumption = extract_ini_primitive_value(ini_data["FUEL"]["CONSUMPTION"], float)
        self.default_fuel = extract_ini_primitive_value(ini_data["FUEL"]["FUEL"], int)
        self.max_fuel = extract_ini_primitive_value(ini_data["FUEL"]["MAX_FUEL"], int)
        self._load_ai_data()
        self.engine = engine.load_engine(self.data_path)
        self.drivetrain = drivetrain.load_drivetrain(os.path.join(self.data_path, "drivetrain.ini"))
        self.shift_lights.load_from_ini(self.data_path)
        self.car_ini_data = ini_data
        self.ui_info.load(self.car_path)

    def swap_engine(self, new_engine, update_mass=False, old_engine_mass=None, use_csp_physics_extensions=False):
        """
        ai.ini:
        [GEARS]
        UP - What RPM AI should shift up
        DOWN - What RPM AI should shift down

        car.ini
        [BASIC] TOTALMASS (Take a value for the chassis and existing engine and adjust accordingly)
        [FUEL] CONSUMPTION
        [INERTIA] If the engine weight changes does this affect this?

        digital_instruments.ini
        [LED_0] - [LED_4] - used for showing shift lights

        drivetrain.ini
        [AUTOCLUTCH]
        [AUTO_SHIFTER]
        [DOWNSHIFT_PROTECTION] (dependant on quality?)


        :param new_engine:
        :param update_mass:
        :param old_engine_mass:
        :param use_csp_physics_extensions
        :return:
        """
        self.using_csp_extended_physics = use_csp_physics_extensions

        if new_engine.metadata.mass_kg and update_mass:
            old_mass = self.engine.metadata.mass_kg if self.engine.metadata.mass_kg else old_engine_mass
            if not old_mass:
                raise CannotUpdateMass()
            self.total_mass -= old_mass
            self.total_mass += new_engine.metadata.mass_kg
        self.engine = new_engine

        if self.using_csp_extended_physics:
            if self.engine.extended_fuel_consumption:
                self.fuel_consumption = None
        else:
            if self.engine.basic_fuel_consumption:
                self.fuel_consumption = self.engine.basic_fuel_consumption

        if not self.data_path:
            return
        self.shift_lights.update(self.engine.limiter)
        if self.engine.metadata.ui_data:
            self.ui_info.torqueCurve = new_engine.metadata.ui_data.torque_curve
            self.ui_info.powerCurve = new_engine.metadata.ui_data.power_curve
            self.ui_info.specs["bhp"] = new_engine.metadata.ui_data.max_power
            self.ui_info.specs["torque"] = new_engine.metadata.ui_data.max_torque
            self.ui_info.specs["weight"] = self.total_mass

    def write(self, output_path=None):
        if output_path is None and self.car_ini_data is None:
            raise IOError("No output file specified")
        ini_data = ac_utils.IniObj(output_path) if output_path else self.car_ini_data
        ini_data.update_attribute("VERSION", self.version, section_name="HEADER")
        ini_data.update_attribute("SCREEN_NAME", self.screen_name, section_name="INFO")
        ini_data.update_attribute("TOTALMASS", self.total_mass, section_name="BASIC")
        if self.using_csp_extended_physics:
            if "CONSUMPTION" in ini_data["FUEL"]:
                ini_data["FUEL"].pop("CONSUMPTION")
        else:
            ini_data.update_attribute("CONSUMPTION", self.fuel_consumption, section_name="FUEL")
        ini_data.update_attribute("FUEL", self.default_fuel, section_name="FUEL")
        ini_data.update_attribute("MAX_FUEL", self.max_fuel, section_name="FUEL")
        ini_data.write()
        self._write_ai_data(ini_data.dirname())
        self.engine.write(ini_data.dirname(), self.using_csp_extended_physics)
        self.drivetrain.write(ini_data.dirname())
        self.shift_lights.write(ini_data.dirname())
        self.ui_info.write(self.car_path if not output_path else output_path)

    def _load_ai_data(self):
        ai_file_path = os.path.join(self.data_path, "ai.ini")
        ai_ini = ac_utils.IniObj(ai_file_path)
        self.ai_shift_up = extract_ini_primitive_value(ai_ini["GEARS"]["UP"], int)
        self.ai_shift_down = extract_ini_primitive_value(ai_ini["GEARS"]["DOWN"], int)

    def _write_ai_data(self, output_dir):
        ai_ini = ac_utils.IniObj(os.path.join(output_dir, "ai.ini"))
        ai_ini.update_attribute("UP", self.ai_shift_up, section_name="GEARS")
        ai_ini.update_attribute("DOWN", self.ai_shift_down, section_name="GEARS")
        ai_ini.write()


class ShiftLights(object):
    def __init__(self):
        self.leds: List[ShiftLED] = list()

    def load_from_ini(self, ini_dir):
        instruments_path = os.path.join(ini_dir, "digital_instruments.ini")
        if not os.path.isfile(instruments_path):
            return
        instruments_ini = ac_utils.IniObj(instruments_path)
        led_idx = 0
        while True:
            led_section_name = f"LED_{led_idx}"
            if led_section_name not in instruments_ini:
                break
            self.leds.append(ShiftLED.create_from_led_section(instruments_ini[led_section_name]))
            led_idx += 1

    def update(self, limiter_rpm):
        # TODO allow spacing of the lights to be provided rather than just 100RPM
        rpm_switch_val = limiter_rpm - 100
        for led in reversed(self.leds):
            led.rpm_switch = rpm_switch_val
            led.blink_switch = limiter_rpm
            rpm_switch_val -= 100

    def write(self, output_path):
        ini_file = ac_utils.IniObj(os.path.join(output_path, "digital_instruments.ini"))
        for led_idx, led in enumerate(self.leds):
            led_section_name = f"LED_{led_idx}"
            if led_section_name not in ini_file:
                ini_file[led_section_name] = dict()
            led.write_to_section(ini_file[led_section_name])
        ini_file.write()


class ShiftLED(object):
    @staticmethod
    def create_from_led_section(section):
        l = ShiftLED()
        l.object_name = section["OBJECT_NAME"]
        l.rpm_switch = section["RPM_SWITCH"]
        l.emissive = section["EMISSIVE"]
        l.diffuse = section["DIFFUSE"]
        l.blink_switch = section["BLINK_SWITCH"]
        l.blink_hz = section["BLINK_HZ"]
        return l

    def __init__(self):
        self.object_name: str = ""
        self.rpm_switch: int = 0
        self.emissive: List[float] = list()
        self.diffuse: float = 0.0
        self.blink_switch: int = 0
        self.blink_hz: int = 0

    def write_to_section(self, section):
        section["OBJECT_NAME"] = self.object_name
        section["RPM_SWITCH"] = self.rpm_switch
        section["EMISSIVE"] = self.emissive
        section["DIFFUSE"] = self.diffuse
        section["BLINK_SWITCH"] = self.blink_switch
        section["BLINK_HZ"] = self.blink_hz


class UIInfo(object):
    def __init__(self):
        self.loaded_ui_json_data = None
        self.ui_json_path = None
        self.name: str = ""
        self.brand: str = ""
        self.description: str = ""
        self.tags: List[str] = list()
        self.car_class: str = ""

        '''
        "specs":
        {
            "bhp": "198bhp",
            "torque": "230Nm",
            "weight": "455kg",
            "topspeed": "230+km/h",
            "acceleration": "--s 0-100",
            "pwratio": "2.30kg/hp",
            "range": 85
        },
        '''
        self.specs = dict()

        self.torqueCurve: List[List[str]] = list()
        self.powerCurve: List[List[str]] = list()

    def load(self, car_path):
        ui_json_path = os.path.join(car_path, os.path.sep.join(["ui", "ui_car.json"]))
        if not os.path.isfile(ui_json_path):
            return
        self.ui_json_path = ui_json_path
        with open(self.ui_json_path, "r") as f:
            self.loaded_ui_json_data = json.load(f)
        for a in ["name", "brand", "description", "tags", "specs", "torqueCurve", "powerCurve"]:
            if a in self.loaded_ui_json_data:
                setattr(self, a, self.loaded_ui_json_data[a])
        if "class" in self.loaded_ui_json_data:
            self.car_class = self.loaded_ui_json_data["class"]

    def write(self, output_dir):
        for a in ["name", "brand", "description", "tags", "specs", "torqueCurve", "powerCurve"]:
            self.loaded_ui_json_data[a] = getattr(self, a)
        self.loaded_ui_json_data["class"] = self.car_class
        with open(os.path.join(output_dir, os.path.sep.join(["ui", "ui_car.json"])), "w+") as f:
            json.dump(self.loaded_ui_json_data, f, indent=4)


"""
    ai.ini:
      [GEARS]
        UP - What RPM AI should shift up
        DOWN - What RPM AI should shift down

    car.ini
      [BASIC] TOTALMASS (Take a value for the chassis and existing engine and adjust accordingly)
      [FUEL] CONSUMPTION
      [INERTIA] If the engine weight changes does this affect this?

    digital_instruments.ini
      [LED_0] - [LED_4] - used for showing shift lights

    drivetrain.ini
      [AUTOCLUTCH]
      [AUTO_SHIFTER]
      [DOWNSHIFT_PROTECTION] (dependant on quality?)

    engine.ini
      [HEADER]
        COAST_CURVE - coast curve. can define 3 different options (coast reference, coast values for mathematical curve, coast curve file)
      [ENGINE_DATA]
        INERTIA
        LIMITER
        MINIMUM (idle rpm)
      [COAST_REF]
        RPM - rev number reference
        TORQUE - engine braking torque value in Nm at rev number reference
        NON_LINEARITY - coast engine brake from ZERO to TORQUE value at rpm with linear (0) to fully exponential (1)
      [TURBO_N]
        LAG_DN - Interpolation lag used slowing down the turbo (0.96 - 0.99)
        LAG_UP - Interpolation lag used to spin up the turbo (0.96 - 0.999)
        MAX_BOOST - Maximum boost generated. This value is never exceeded and multiply the torque
                        like T=T*(1.0 + boost), a boost of 2 will give you 3 times the torque at a given rpm.
        WASTEGATE - Max level of boost before the wastegate does its things. 0 = no wastegate
        DISPLAY_MAX_BOOST - Value used by display apps
        REFERENCE_RPM= - The reference rpm where the turbo reaches maximum boost (at max gas pedal)
        GAMMA=5 - A value used to make the boost curve more exponential. 1 = linear
        COCKPIT_ADJUSTABLE=0

      fuel_cons.ini
        [FUEL_EVAL]
          KM_PER_LITER

      power.lut
        A lookup table for torque values after applying drivetrain losses. Each line is <RPM>|<TORQUE_NM>

      sounds.ini
      No idea how to update these yet. An example from tatuusfa01
      [BACKFIRE]
      MAXGAS=0.4
      MINRPM=3500
      MAXRPM=15000
      TRIGGERGAS=0.8
      VOLUME_IN=1.0
      VOLUME_OUT=0.5
      VOLUME_SCALE_OUT=6

    Warnings:
      if engine torque is greater than [CLUTCH].MAX_TORQUE in drivetrain.ini
      
    Update pretty much everything in drivetrain.ini

    Create ratios.rto files for the gears available for the gearbox
    The files are of the form <ratio-name>|<ratio-value> e.g: F3B-12:38-INT|3.17

    Reference these ratio files in setup.ini
    [GEAR_N] where N is the gear number
      RATIOS=<ratios-file>
    [FINAL_GEAR_RATIO]
      RATIOS=<final.rto>

    In drivetrain.ini
    [GEARS]
     COUNT, GEAR_R, FINAL, GEAR_N (must be a gear for 1-N where N is COUNT)
     these should match up with a ratio lookup table file
"""



import setuptools

version = {}
with open("sim_racing_tools/version.py") as fp:
    exec(fp.read(), version)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sim-racing-tools",
    keywords=["Assetto Corsa", "Automation", "Sim Racing"],
    version=version['__version__'],
    author="zephyrj",
    author_email="zephyrj@protonmail.com",
    description="A collection of tools to make sim-racing more fun",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zephyrj/sim-racing-tools",
    download_url="https://github.com/zephyrj/sim-racing-tools/archive/refs/tags/2021.6.2.tar.gz",
    project_urls={
        "Bug Tracker": "https://github.com/zephyrj/sim-racing-tools/issues",
    },
    packages=setuptools.find_packages('.'),
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=["setuptools~=54.2.0",
                      "wheel~=0.36.2",
                      "toml~=0.10.2",
                      "six~=1.15.0",
                      "configobj~=5.0.6",
                      "argcomplete~=1.12.3",
                      "pandas~=1.2.4",
                      "plotly~=4.14.3",
                      "pywin32"],
    scripts=[],
    entry_points={
        'console_scripts': ['ac-tools=sim_racing_tools.assetto_corsa.scripts.ac_tools:main',
                            'check-engine=sim_racing_tools.automation.scripts.check_engine:main'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ]
)



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys

import sim_racing_tools.constants as constants


GAME_NAME = "assettocorsa"
GAME_ID = 244210

CARS_PATH = os.sep.join(["content", "cars"])
SFX_PATH = os.sep.join(["content", "sfx"])

DATA_FOLDER_NAME = "data"
ENCODED_DATA_FILENAME = "data.acd"
SFX_GUID_FILENAME = "GUIDs.txt"


def get_install_dir():
    return os.path.join(constants.get_game_install_path(), GAME_NAME)


def get_cars_dir():
    return Installation().get_installed_cars_path()


class Installation(object):
    def __init__(self, custom_root=None):
        self.on_linux = sys.platform == "linux" or sys.platform == "linux2"
        self.installation_root = get_install_dir() if not custom_root else custom_root
        self._build_installed_car_map()
        self._build_sfx_guid_map()

    def get_installed_cars_path(self):
        return os.path.join(self.installation_root, CARS_PATH)

    def _build_installed_car_map(self):
        self.installed_cars = set()
        _, dirnames, _ = next(os.walk(self.get_installed_cars_path()))
        self.installed_cars = {name for name in dirnames}

    def _build_sfx_guid_map(self):
        self.sfx_dict_by_folder_name = dict()
        self.sfx_bank_dict = dict()
        root_gui_file_path = os.path.join(self.installation_root,
                                          os.sep.join([SFX_PATH, SFX_GUID_FILENAME]))
        with open(root_gui_file_path, "r") as main_sfx_guid_file:
            for line in main_sfx_guid_file:
                line_data = line.split()
                guid = line_data[0]
                sfx_line = line_data[1]
                if sfx_line.startswith("event"):
                    folder_name = sfx_line.split(":")[1].split("/")[2]
                    if folder_name not in self.sfx_dict_by_folder_name:
                        self.sfx_dict_by_folder_name[folder_name] = list()
                    self.sfx_dict_by_folder_name[folder_name].append(line)
                elif sfx_line.startswith("bank"):
                    self.sfx_bank_dict[sfx_line.split("/")[1]] = guid

    def _generate_data_dir(self, new_car_path, existing_car_acd_path):
        # todo
        pass



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import struct
import toml
import logging

from collections import OrderedDict

from typing import List

# Appears to denote the start of a blob
BLOB_MARK = 1
# Bytes denoting true or false
FALSE = 48
TRUE = 49
# A number seems to represent a REAL or a NUMBER but is always stored as 8 byte Double
NUMBER = 78
# TEXT can be a string or a blob of data
TEXT = 83
# This tag denote other attributes live inside this one
ATTRIBUTE_SECTION = 84


def type_byte_to_string(byte):
    if byte == NUMBER:
        return "Number"
    elif byte == TEXT:
        return "Text"
    elif byte == ATTRIBUTE_SECTION:
        return "AttributeSection"
    return str(byte)


class ByteChunk(object):
    def __init__(self, byte_stream):
        self.byte_stream = byte_stream

    def hex(self):
        return " ".join(f'{b:02X}' for b in self.byte_stream)

    @property
    def bytes(self):
        return self.byte_stream

    @property
    def length(self):
        return len(self.byte_stream)


class AttributeSection(ByteChunk):
    def __init__(self, name, byte_stream):
        super(AttributeSection, self).__init__(byte_stream)
        self.name = name
        if struct.unpack("<I", bytearray(byte_stream[0:4]))[0] > 0:
            self.num_children = struct.unpack("<LL", bytearray(byte_stream[0:8]))[0]
        else:
            self.num_children = struct.unpack("<I", bytearray(byte_stream[4:8]))[0]
        self.attribute_list = list()
        self.section_list = list()
        self.section_stack = list()

    def __str__(self):
        return "\n".join(self.summary())

    def as_dict(self):
        data_dict = OrderedDict()
        for attribute in self.attribute_list:
            data_dict[attribute.name] = attribute.value
        for section in self.section_list:
            data_dict[section.name] = section.as_dict()
        return data_dict

    def summary(self):
        lines = list()
        lines.append(f'{self.name}')
        lines.extend([f'  {str(a)}' for a in self.attribute_list])
        for section in self.section_list:
            lines.extend([f'  {line}' for line in section.summary()])
            lines.append("")
        return lines

    def add_attribute(self, attribute):
        if len(self.section_stack):
            self.section_stack[-1].add_attribute(attribute)
        else:
            self.attribute_list.append(attribute)
        self._check_for_complete_sections()

    def add_section(self, section):
        if len(self.section_stack):
            self.section_stack[-1].add_section(section)
        else:
            self.section_stack.append(section)
            self.section_list.append(self.section_stack[-1])
        self._check_for_complete_sections()

    def complete(self):
        if len(self.section_stack):
            return False
        return self.num_children == (len(self.attribute_list) + len(self.section_list))

    def _check_for_complete_sections(self):
        while len(self.section_stack) and self.section_stack[-1].complete():
            self.section_stack.pop()


class Blob(ByteChunk):
    def __init__(self, name, byte_stream):
        super(Blob, self).__init__(byte_stream)
        self.name = name
        self.attributes = list()
        self.sections = list()
        self.section_stack = list()


class Text(ByteChunk):
    def __init__(self, byte_stream):
        super(Text, self).__init__(byte_stream)

    def __str__(self):
        try:
            return self.byte_stream.decode("utf-8")
        except (AttributeError, ValueError):
            return self.hex()

    @property
    def value(self):
        return self.__str__()


class Number(ByteChunk):
    def __init__(self, byte_stream):
        super(Number, self).__init__(byte_stream)
        self._value = struct.unpack("d", bytearray(self.byte_stream))[0]

    def __str__(self):
        return str(self._value)

    @property
    def value(self):
        return self._value


class FalseType(ByteChunk):
    def __init__(self, byte_stream):
        super(FalseType, self).__init__(byte_stream)

    def __str__(self):
        return str(False)

    @property
    def value(self):
        return False


class TrueType(ByteChunk):
    def __init__(self, byte_stream):
        super(TrueType, self).__init__(byte_stream)

    def __str__(self):
        return str(True)

    @property
    def value(self):
        return True


class CarAttribute(object):
    def __init__(self):
        self.name = None
        self.value_object = None

    def __str__(self):
        return f'{self.name} = {str(self.value_object)}'

    @property
    def value(self):
        return self.value_object.value


class CarFile(object):
    def __init__(self, filename, byte_stream):
        attributes: List[CarAttribute]
        sections: List[AttributeSection]

        self.car_file_path = filename
        self.byte_stream = byte_stream
        self.current_pos = 0
        self.section_stack = list()
        self.sections = list()
        self.attributes = list()
        self.attribute_offset_map = dict()
        self._parse_opening_blob_mark("Car")
        self.current_attribute = None

    def __str__(self):
        out = "\n\n".join([str(a) for a in self.attributes])
        for section in self.sections:
            out += str(section)
        return out

    def get_data(self):
        data_dict = OrderedDict()
        data_dict["car-file-path"] = self.car_file_path

        for attribute in self.attributes:
            data_dict[attribute.name] = attribute.value

        for section in self.sections:
            data_dict[section.name] = section.as_dict()
        return data_dict

    def write_toml(self, path):
        with open(path, "w+") as out_file:
            toml.dump(self.get_data(), out_file)

    def parse(self):
        parsing_attribute = False
        parsing_int_pair = False
        while True:
            if self.current_pos >= len(self.byte_stream):
                break
            if parsing_attribute or parsing_int_pair:
                if not parsing_int_pair:
                    length = self._parse_length()
                    if self._is_blob():
                        logging.debug("Skipped non-string:")
                        logging.debug(" ".join(f'{b:02X}' for b in self.byte_stream[self.current_pos: self.current_pos + length]))
                        self.current_pos += length
                        continue
                    attribute_name = self._parse_text(length)
                else:
                    attribute_name = str(self._parse_number())

                object_type = self._parse_type()
                if object_type == ATTRIBUTE_SECTION:
                    self._add_section(self._parse_attribute_section(str(attribute_name)))
                elif object_type == FALSE:
                    self._add_attribute(str(attribute_name), FalseType(object_type))
                elif object_type == TRUE:
                    self._add_attribute(str(attribute_name), TrueType(object_type))
                else:
                    if object_type == NUMBER:
                        self._add_attribute(str(attribute_name), self._parse_number())
                    else:
                        length = self._parse_length()
                        if self._is_blob():
                            self.current_pos += 2
                            self._add_section(self._parse_attribute_section(str(attribute_name)))
                        else:
                            self._add_attribute(str(attribute_name), self._parse_text(length))

                self._check_for_section_complete()
                parsing_attribute = False
                parsing_int_pair = False
            else:
                type_byte = self._parse_type()
                if type_byte == TEXT:
                    parsing_attribute = True
                elif type_byte == NUMBER:
                    parsing_int_pair = True

    def _check_for_section_complete(self):
        while len(self.section_stack) and self.section_stack[-1].complete():
            self.section_stack.pop()

    def _add_section(self, section):
        if len(self.section_stack):
            self.section_stack[-1].add_section(section)
        else:
            self.section_stack.append(section)
            self.sections.append(self.section_stack[-1])

    def _add_attribute(self, name, value):
        attribute = CarAttribute()
        attribute.name = name
        attribute.value_object = value

        if len(self.section_stack):
            self.section_stack[-1].add_attribute(attribute)
        else:
            self.attributes.append(attribute)
        self.attribute_offset_map[attribute.name] = self.current_pos

    def _is_blob(self):
        if self.byte_stream[self.current_pos] == 1:
            return True
        return False

    def _parse_string_length(self):
        try:
            return struct.unpack("<I", bytearray(self.byte_stream[self.current_pos:self.current_pos + 4]))[0]
        finally:
            self.current_pos += 4

    def _parse_text(self, length):
        try:
            return Text(self.byte_stream[self.current_pos:self.current_pos + length])
        finally:
            self.current_pos += length

    def _parse_number(self):
        try:
            return Number(self.byte_stream[self.current_pos:self.current_pos + 8])
        finally:
            self.current_pos += 8

    def _parse_attribute_section(self, name):
        try:
            return AttributeSection(name, self.byte_stream[self.current_pos:self.current_pos + 8])
        finally:
            self.current_pos += 8

    def _parse_type(self):
        try:
            return self.byte_stream[self.current_pos]
        finally:
            self.current_pos += 1

    def _parse_opening_blob_mark(self, opening_section_name):
        if self.byte_stream[self.current_pos] != BLOB_MARK:
            raise ValueError("File doesn't open with a blob mark - is it a valid .car file?")
        self.current_pos += 2  # Jump over
        self.section_stack.append(self._parse_attribute_section(opening_section_name))
        self.sections.append(self.section_stack[-1])

    def _parse_length(self):
        try:
            return struct.unpack("<I", bytearray(self.byte_stream[self.current_pos:self.current_pos + 4]))[0]
        finally:
            self.current_pos += 4


def hp_lut_calculator(torque_lut):
    # hp = T x RPM/5,252
    out_lut = OrderedDict()
    for rpm, T in torque_lut:
        out_lut[rpm] = T * rpm/5252


if __name__ == '__main__':
    with open(sys.argv[1], "rb") as f:
        car_bytes = f.read()
        c = CarFile(f.name, car_bytes)
        c.parse()
        c.write_toml("out.toml")



"""
Copyright (c):
2021 zephyrj
zephyrj@protonmail.com

This file is part of sim-racing-tools.

sim-racing-tools is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

sim-racing-tools is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with sim-racing-tools. If not, see <https://www.gnu.org/licenses/>.
"""

# TODO ensure we install quickbms and the assetto corsa plugin
# TODO Create a data folder beforehand so that we don't get a prompt
QUICKBMS_EXE_NAME = "quickbms.exe"



from steam.steamid import SteamID


def install_epsxe():
    pass


def add_game_to_steam(steam_id=None, steam_account_number=None):
    if not steam_id and not steam_account_number:
        raise ValueError("Must provide steam_id or steam_partner_id")

    if steam_id:
        folder_name = str(SteamID(steam_id) & 0xFFFFFFFE)
    else:
        folder_name = str(steam_account_number)


